from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Type
from uuid import UUID

from django.apps import apps as django_apps
from django.utils.translation import gettext as _
from edc_utils import get_utcnow
from edc_view_utils.dashboard_model_button import DashboardModelButton
from edc_view_utils.model_button import ADD

from ..constants import IN_PROGRESS_APPT, NEW_APPT

if TYPE_CHECKING:
    from ..models import Appointment


__all__ = ["AppointmentButton"]


@dataclass
class AppointmentButton(DashboardModelButton):
    model_obj: Appointment = None
    colors: tuple[str, str, str] = field(default=(3 * ("default",)))
    model_cls: Type[Appointment] = field(default=None, init=False)
    appointment: Appointment = field(default=None, init=False)

    def __post_init__(self):
        self.model_cls = django_apps.get_model("edc_appointment.appointment")

    @property
    def disabled(self) -> str:
        disabled = "disabled"
        if self.model_obj.appt_status == IN_PROGRESS_APPT and self.perms.change:
            disabled = ""
        elif (
            self.model_obj.appt_status != NEW_APPT
            and not self.perms.add
            and not self.perms.change
            and self.perms.view
        ):
            disabled = ""
        return disabled

    @property
    def label(self) -> str:
        return _("Appt")

    @property
    def color(self) -> str:
        color = super().color
        if (
            self.model_obj
            and self.model_obj.appt_datetime <= get_utcnow()
            and not self.model_obj.related_visit
        ):
            color = self.colors[ADD]
        return color

    @property
    def fa_icon(self) -> str:
        fa_icon = super().fa_icon
        if self.model_obj.appt_status == NEW_APPT:
            fa_icon = self.fa_icons[ADD]
        return fa_icon

    @property
    def reverse_kwargs(self) -> dict[str, str | UUID]:
        return dict(
            subject_identifier=self.model_obj.subject_identifier,
        )

    @property
    def extra_kwargs(self) -> dict[str, str | int | UUID]:
        return dict(reason=self.model_obj.appt_reason)



from __future__ import annotations

import datetime as dt
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings, tag
from edc_consent import site_consents
from edc_facility.import_holidays import import_holidays
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from edc_appointment.constants import INCOMPLETE_APPT
from edc_appointment.models import Appointment
from edc_appointment_app.consents import consent_v1
from edc_appointment_app.models import SubjectVisit
from edc_appointment_app.tests.appointment_app_test_case_mixin import (
    AppointmentAppTestCaseMixin,
)
from edc_appointment_app.visit_schedule import get_visit_schedule3

from ...creators import UnscheduledAppointmentCreator
from ..helper import Helper

utc_tz = ZoneInfo("UTC")


test_datetime = dt.datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz)


@override_settings(SITE_ID=10)
@time_machine.travel(test_datetime)
class TestInsertUnscheduled(AppointmentAppTestCaseMixin, TestCase):
    helper_cls = Helper

    def setUp(self):
        self.subject_identifier = "12345"

        site_consents.registry = {}
        site_consents.register(consent_v1)

        site_visit_schedules._registry = {}
        self.visit_schedule = get_visit_schedule3(consent_v1)
        site_visit_schedules.register(self.visit_schedule)

        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=test_datetime,
        )

        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule.name,
            schedule_name="three_monthly_schedule",
        )
        appointments = Appointment.objects.filter(
            subject_identifier=self.subject_identifier,
        )
        self.assertEqual(appointments.count(), 5)

        appointment = Appointment.objects.get(timepoint=0.0)
        self.create_related_visit(appointment)
        appointment = Appointment.objects.get(timepoint=3.0)
        self.create_related_visit(appointment)
        appointment = Appointment.objects.get(timepoint=6.0)
        self.create_related_visit(appointment)
        appointment = Appointment.objects.get(timepoint=9.0)
        self.create_related_visit(appointment)
        appointment = Appointment.objects.get(timepoint=12.0)
        self.create_related_visit(appointment)

        self.appt_datetimes = [
            o.appt_datetime for o in Appointment.objects.all().order_by("appt_datetime")
        ]

    @classmethod
    def setUpTestData(cls):
        import_holidays()

    @staticmethod
    def create_unscheduled(appointment: Appointment, days: int = None):
        creator = UnscheduledAppointmentCreator(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            suggested_appt_datetime=appointment.appt_datetime + relativedelta(days=days),
            suggested_visit_code_sequence=appointment.visit_code_sequence + 1,
        )
        appointment = creator.appointment
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save_base(update_fields=["appt_status"])
        return appointment

    @staticmethod
    def get_visit_codes(by: str = None, visit_schedule_name: str | None = None, **kwargs):
        opts = dict(visit_schedule_name=visit_schedule_name)
        return [
            f"{o.visit_code}.{o.visit_code_sequence}"
            for o in Appointment.objects.filter(**opts).order_by(by)
        ]

    @tag("5")
    def test_insert_unscheduled_between_related_visits(self):

        appt1030 = Appointment.objects.get(
            visit_code="1030",
            visit_code_sequence=0,
            visit_schedule_name=self.visit_schedule.name,
        )

        appt = self.create_unscheduled(appt1030, days=2)
        self.create_related_visit(appt)
        appt = self.create_unscheduled(appt1030, days=3)
        self.create_related_visit(appt)
        appt = self.create_unscheduled(appt1030, days=4)
        self.create_related_visit(appt)
        appt = self.create_unscheduled(appt1030, days=6)
        self.create_related_visit(appt)
        appt = self.create_unscheduled(appt1030, days=7)
        self.create_related_visit(appt)
        appt = self.create_unscheduled(appt1030, days=8)
        self.create_related_visit(appt)

        visit_codes = self.get_visit_codes(
            by="appt_datetime",
            visit_schedule_name=self.visit_schedule.name,
        )

        # appointments
        self.assertEqual(
            [
                "1000.0",
                "1030.0",
                "1030.1",
                "1030.2",
                "1030.3",
                "1030.4",
                "1030.5",
                "1030.6",
                "1060.0",
                "1090.0",
                "1120.0",
            ],
            visit_codes,
        )

        # related subject_visit
        self.assertEqual(
            [
                "1000.0",
                "1030.0",
                "1030.1",
                "1030.2",
                "1030.3",
                "1030.4",
                "1030.5",
                "1030.6",
                "1060.0",
                "1090.0",
                "1120.0",
            ],
            [
                f"{o.visit_code}.{o.visit_code_sequence}"
                for o in Appointment.objects.filter(
                    visit_schedule_name=self.visit_schedule.name,
                ).order_by("appt_datetime")
                if getattr(o, "related_visit", None)
            ],
        )

        # insert an unscheduled appointment between 1030.3 and 1030.4
        self.create_unscheduled(appt1030, days=5)

        # appointments
        visit_codes = self.get_visit_codes(
            by="appt_datetime",
            visit_schedule_name=self.visit_schedule.name,
        )
        self.assertEqual(
            [
                "1000.0",
                "1030.0",
                "1030.1",
                "1030.2",
                "1030.3",
                "1030.4",
                "1030.5",
                "1030.6",
                "1030.7",
                "1060.0",
                "1090.0",
                "1120.0",
            ],
            visit_codes,
        )

        # related subject_visit
        self.assertEqual(
            [
                "1000.0: 1000.0",
                "1030.0: 1030.0",
                "1030.1: 1030.1",
                "1030.2: 1030.2",
                "1030.3: 1030.3",
                "1030.5: 1030.5",
                "1030.6: 1030.6",
                "1030.7: 1030.7",
                "1060.0: 1060.0",
                "1090.0: 1090.0",
                "1120.0: 1120.0",
            ],
            [
                f"{o.appointment.visit_code}.{o.appointment.visit_code_sequence}: "
                f"{o.visit_code}.{o.visit_code_sequence}"
                for o in SubjectVisit.objects.filter(
                    visit_schedule_name=self.visit_schedule.name,
                ).order_by("report_datetime")
            ],
        )



from dateutil.relativedelta import relativedelta
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.visit import Visit
from edc_visit_schedule.visit_schedule import VisitSchedule

from edc_appointment_app.consents import consent_v1
from edc_appointment_app.visit_schedule.crfs import crfs, crfs_missed


def get_visit_schedule4() -> VisitSchedule:
    visit_schedule4 = VisitSchedule(
        name="visit_schedule4",
        offstudy_model="edc_appointment_app.subjectoffstudy",
        death_report_model="edc_appointment_app.deathreport",
    )

    schedule4 = Schedule(
        name="three_monthly_schedule",
        onschedule_model="edc_appointment_app.onschedulethree",
        offschedule_model="edc_appointment_app.offschedulethree",
        appointment_model="edc_appointment.appointment",
        consent_definitions=[consent_v1],
    )

    visits = [
        Visit(
            code="1000",
            title="Baseline",
            timepoint=0,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=0),
            add_window_gap_to_lower=True,
            requisitions=None,
            crfs=crfs,
            crfs_missed=crfs_missed,
            facility_name="7-day-clinic",
        )
    ]
    for index, visit_code in [(3, "1030"), (6, "1060"), (9, "1090"), (12, "1120")]:
        visits.append(
            Visit(
                code=visit_code,
                title=f"Month {index}",
                timepoint=index,
                rbase=relativedelta(months=index),
                rlower=relativedelta(days=15),
                rupper=relativedelta(days=15),
                add_window_gap_to_lower=True,
                max_window_gap_to_lower=None,
                requisitions=None,
                crfs=crfs,
                crfs_missed=crfs_missed,
                facility_name="7-day-clinic",
                allow_unscheduled=True,
            )
        )
    for visit in visits:
        schedule4.add_visit(visit)

    visit_schedule4.add_schedule(schedule4)
    return visit_schedule4



from __future__ import annotations

import calendar
from typing import TYPE_CHECKING, Any

from django.contrib import admin
from django.db.models import DurationField, ExpressionWrapper, F
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext as _
from django_audit_fields.admin import audit_fieldset_tuple
from edc_constants.constants import NOT_APPLICABLE
from edc_document_status.fieldsets import document_status_fieldset_tuple
from edc_document_status.modeladmin_mixins import DocumentStatusModelAdminMixin
from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin
from edc_model_admin.history import SimpleHistoryAdmin
from edc_sites.admin import SiteModelAdminMixin
from edc_utils import get_utcnow
from edc_visit_schedule.admin import ScheduleStatusListFilter
from edc_visit_schedule.exceptions import OnScheduleError
from edc_visit_schedule.fieldsets import (
    visit_schedule_fields,
    visit_schedule_fieldset_tuple,
)
from edc_visit_schedule.utils import off_schedule_or_raise

from ..admin_site import edc_appointment_admin
from ..choices import APPT_STATUS, APPT_TIMING, DEFAULT_APPT_REASON_CHOICES
from ..constants import NEW_APPT, SKIPPED_APPT
from ..forms import AppointmentForm
from ..models import Appointment, AppointmentType
from ..utils import get_allow_skipped_appt_using
from .actions import appointment_mark_as_done, appointment_mark_as_new
from .list_filters import (
    AppointmentListFilter,
    AppointmentOverdueListFilter,
    AppointmentStatusListFilter,
)

if TYPE_CHECKING:
    from django.db.models import QuerySet


@admin.register(Appointment, site=edc_appointment_admin)
class AppointmentAdmin(
    SiteModelAdminMixin,
    ModelAdminSubjectDashboardMixin,
    DocumentStatusModelAdminMixin,
    SimpleHistoryAdmin,
):
    show_cancel = True
    form = AppointmentForm
    actions = [appointment_mark_as_done, appointment_mark_as_new]
    date_hierarchy = "appt_datetime"
    list_display = (
        "appointment_subject",
        "full_visit_code",
        "appt_actions",
        "appointment_date",
        "appt_status",
        "days_from_timepoint_datetime",
        "days_from_now",
        "appointment_type",
        "timing",
        "schedule_name",
    )
    list_filter = (
        AppointmentListFilter,
        AppointmentStatusListFilter,
        "visit_code",
        "visit_code_sequence",
        "appt_type",
        "appt_timing",
        ScheduleStatusListFilter,
        AppointmentOverdueListFilter,
    )

    additional_instructions = format_html(
        "{}.<BR><i>{}.</i>",
        (
            "To start or continue to edit FORMS for this subject, change the "
            "appointment status below to 'In Progress' and click SAVE"
        ),
        (
            "Note: You may only edit one appointment at a time. "
            "Before you move to another appointment, change the appointment "
            "status below to 'Incomplete' or 'Done'"
        ),
    )

    fieldsets = (
        (
            None,
            (
                {
                    "fields": (
                        "subject_identifier",
                        "appt_datetime",
                        "appt_type",
                        "appt_status",
                        "appt_reason",
                        "appt_timing",
                        "comment",
                    )
                }
            ),
        ),
        (
            "Timepoint",
            (
                {
                    "classes": ("collapse",),
                    "fields": (
                        "timepoint",
                        "timepoint_datetime",
                        "visit_code_sequence",
                        "facility_name",
                    ),
                }
            ),
        ),
        document_status_fieldset_tuple,
        visit_schedule_fieldset_tuple,
        audit_fieldset_tuple,
    )

    radio_fields = {
        "appt_type": admin.VERTICAL,
        "appt_status": admin.VERTICAL,
        "appt_reason": admin.VERTICAL,
        "appt_timing": admin.VERTICAL,
    }

    search_fields = ("subject_identifier",)

    def get_readonly_fields(self, request, obj=None) -> tuple:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        return (
            readonly_fields
            + visit_schedule_fields
            + (
                "subject_identifier",
                "timepoint",
                "timepoint_datetime",
                "visit_code_sequence",
                "facility_name",
            )
        )

    def has_delete_permission(self, request, obj=None):
        """Override to remove delete permissions if OnSchedule
        and visit_code_sequence == 0.

        See `edc_visit_schedule.off_schedule_or_raise()`
        """
        has_delete_permission = super().has_delete_permission(request, obj=obj)
        if has_delete_permission and obj:
            if obj.visit_code_sequence == 0 or (
                obj.visit_code_sequence != 0 and obj.appt_status != NEW_APPT
            ):
                try:
                    off_schedule_or_raise(
                        subject_identifier=obj.subject_identifier,
                        report_datetime=obj.appt_datetime,
                        visit_schedule_name=obj.visit_schedule_name,
                        schedule_name=obj.schedule_name,
                    )
                except OnScheduleError:
                    has_delete_permission = False
        return has_delete_permission

    @admin.display(description="Timing", ordering="appt_timing")
    def timing(self, obj=None):
        if obj.appt_status == NEW_APPT:
            return None
        return obj.get_appt_timing_display()

    @admin.display(description="Visit", ordering="visit_code")
    def full_visit_code(self, obj=None):
        """Returns a string of visit_code.visit_code_sequence"""
        return f"{obj.visit_code}.{obj.visit_code_sequence}"

    @admin.display(description="Appt. Date", ordering="appt_datetime")
    def appointment_date(self, obj=None):
        return f"{obj.appt_datetime.date()} {calendar.day_abbr[obj.appt_datetime.weekday()]}"

    @admin.display(description="Type", ordering="appt_type")
    def appointment_type(self, obj=None):
        return obj.get_appt_type_display()

    @admin.display(description="Timepoint date", ordering="timepoint_datetime")
    def timepoint_date(self, obj=None):
        timepoint_date = obj.timepoint_datetime.date()
        weekday = calendar.day_abbr[obj.timepoint_datetime.weekday()]
        return f"{timepoint_date} {weekday}"

    @admin.display(description="Timepoint", ordering="appt_timepoint_delta")
    def days_from_timepoint_datetime(self, obj=None):
        if obj.appt_datetime.time() >= obj.timepoint_datetime.time():
            days = obj.appt_timepoint_delta.days
        else:
            days = obj.appt_timepoint_delta.days + 1
        if days == 0:
            return None
        return f"{'+' if days > 0 else ''}{days}d"

    @admin.display(description="Now", ordering="appt_datetime_delta")
    def days_from_now(self, obj=None):
        days = obj.appt_datetime_delta.days
        if days == 0:
            return None
        return f"{'+' if days > 0 else ''}{days}d"

    @admin.display(description="Subject", ordering="subject_identifier")
    def appointment_subject(self, obj=None):
        return obj.subject_identifier

    @admin.display(description="Options")
    def appt_actions(self, obj=None):
        dashboard_url = reverse(
            self.get_subject_dashboard_url_name(),
            kwargs=self.get_subject_dashboard_url_kwargs(obj),
        )
        call_url = "#"
        context = dict(
            dashboard_title=_("Go to subject's dashboard"),
            dashboard_url=dashboard_url,
            call_title=_("Call subject"),
            call_url=call_url,
        )
        return render_to_string("button.html", context=context)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "appt_type":
            kwargs["queryset"] = self.get_appt_type_queryset(request)
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def formfield_for_choice_field(self, db_field, request, **kwargs):
        if db_field.name == "appt_reason":
            kwargs["choices"] = self.get_appt_reason_choices(request)
        if db_field.name == "appt_status":
            kwargs["choices"] = self.get_appt_status_choices(request)
        if db_field.name == "appt_timing":
            kwargs["choices"] = self.get_appt_timing_choices(request)
        return super().formfield_for_choice_field(db_field, request, **kwargs)

    def get_appt_type_queryset(self, request) -> QuerySet:
        if not self.allow_skipped_appointments(request):
            return AppointmentType.objects.exclude(name=NOT_APPLICABLE).order_by(
                "display_index"
            )
        return AppointmentType.objects.all().order_by("display_index")

    def get_appt_reason_choices(self, request) -> tuple[Any, ...]:
        """Return a choices tuple.

        Important: left side of the tuple MUST have the default
        values of SCHEDULED_APPT and UNSCHEDULED_APPT.
        """
        return DEFAULT_APPT_REASON_CHOICES

    def get_appt_status_choices(self, request) -> tuple[Any, ...]:
        if not self.allow_skipped_appointments(request):
            return tuple([tpl for tpl in APPT_STATUS if tpl[0] != SKIPPED_APPT])
        return APPT_STATUS

    def get_appt_timing_choices(self, request) -> tuple[Any, ...]:
        if not self.allow_skipped_appointments(request):
            return tuple([tpl for tpl in APPT_TIMING if tpl[0] != NOT_APPLICABLE])
        return APPT_TIMING

    def allow_skipped_appointments(self, request) -> bool:
        """Returns True if settings.EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING
        has value.

        Relates to use of `SKIPPED_APPT` feature.
        """
        return True if get_allow_skipped_appt_using() else False

    def get_queryset(self, request):
        now = get_utcnow().replace(second=59, hour=23, minute=59)
        return (
            super()
            .get_queryset(request)
            .annotate(
                appt_timepoint_delta=ExpressionWrapper(
                    (F("appt_datetime") - F("timepoint_datetime")),
                    output_field=DurationField(),
                ),
                appt_datetime_delta=ExpressionWrapper(
                    (F("appt_datetime") - now), output_field=DurationField()
                ),
            )
        )



from __future__ import annotations

from edc_form_runners.decorators import register
from edc_form_runners.form_runner import FormRunner


@register()
class AppointmentFormRunner(FormRunner):
    model_name = "edc_appointment.appointment"
    extra_fieldnames = ["appt_datetime"]
    exclude_formfields = ["appt_close_datetime"]



from typing import Type

from edc_consent import site_consents
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED

from edc_appointment.constants import INCOMPLETE_APPT
from edc_appointment.creators import UnscheduledAppointmentCreator
from edc_appointment.models import Appointment
from edc_appointment_app.consents import consent_v1
from edc_appointment_app.models import SubjectVisit
from edc_appointment_app.visit_schedule import (
    get_visit_schedule1,
    get_visit_schedule2,
    get_visit_schedule3,
)


class AppointmentAppTestCaseMixin:
    helper_cls = None

    def setUp(self):
        super().setUp()

        self.subject_identifier = "12345"
        site_visit_schedules._registry = {}
        self.visit_schedule1 = get_visit_schedule1()
        self.visit_schedule2 = get_visit_schedule2()
        self.visit_schedule3 = get_visit_schedule3()
        site_visit_schedules.register(self.visit_schedule1)
        site_visit_schedules.register(self.visit_schedule2)
        site_visit_schedules.register(self.visit_schedule3)
        site_consents.registry = {}
        site_consents.register(consent_v1)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=ResearchProtocolConfig().study_open_datetime,
        )
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule1.name, schedule_name="schedule1"
        )
        appointments = Appointment.objects.filter(subject_identifier=self.subject_identifier)
        self.assertEqual(appointments.count(), 4)

        appointment = Appointment.objects.get(timepoint=0.0)
        self.create_related_visit(appointment)
        self.create_unscheduled_appointments(appointment)

        self.appt_datetimes = [
            o.appt_datetime for o in Appointment.objects.all().order_by("appt_datetime")
        ]

    def create_unscheduled_appointments(self, appointment):
        for i in range(0, 3):
            creator = UnscheduledAppointmentCreator(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule1.name,
                schedule_name="schedule1",
                visit_code="1000",
                suggested_visit_code_sequence=appointment.visit_code_sequence + 1,
            )
            appointment = creator.appointment
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save_base(update_fields=["appt_status"])

    @staticmethod
    def create_related_visit(
        appointment: Appointment, reason: str | None = None
    ) -> SubjectVisit:
        if not appointment.related_visit:
            related_visit = SubjectVisit.objects.create(
                appointment=appointment,
                subject_identifier=appointment.subject_identifier,
                report_datetime=appointment.appt_datetime,
                visit_schedule_name=appointment.visit_schedule_name,
                schedule_name=appointment.schedule_name,
                visit_code=appointment.visit_code,
                visit_code_sequence=appointment.visit_code_sequence,
                reason=reason or SCHEDULED,
            )
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save_base(update_fields=["appt_status"])
            appointment.refresh_from_db()
        else:
            related_visit = appointment.related_visit
        return related_visit

    @staticmethod
    def get_visit_codes(model_cls: Type[Appointment | SubjectVisit], order_by: str = None):
        return [
            f"{o.visit_code}.{o.visit_code_sequence}"
            for o in model_cls.objects.all().order_by(order_by)
        ]



from dateutil.relativedelta import relativedelta
from edc_consent.consent_definition import ConsentDefinition
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.visit import Visit
from edc_visit_schedule.visit_schedule import VisitSchedule

from ..consents import consent_v1
from .crfs import crfs, crfs_missed, requisitions


def get_visit_schedule3(cdef: ConsentDefinition | None = None) -> VisitSchedule:
    visit_schedule3 = VisitSchedule(
        name="visit_schedule3",
        offstudy_model="edc_appointment_app.subjectoffstudy",
        death_report_model="edc_appointment_app.deathreport",
    )

    schedule3 = Schedule(
        name="three_monthly_schedule",
        onschedule_model="edc_appointment_app.onschedulethree",
        offschedule_model="edc_appointment_app.offschedulethree",
        appointment_model="edc_appointment.appointment",
        consent_definitions=[cdef or consent_v1],
    )

    visits = [
        Visit(
            code="1000",
            title="Baseline",
            timepoint=0,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=0),
            requisitions=requisitions,
            crfs=crfs,
            crfs_missed=crfs_missed,
            facility_name="7-day-clinic",
        )
    ]
    for index, visit_code in [(3, "1030"), (6, "1060"), (9, "1090"), (12, "1120")]:
        visits.append(
            Visit(
                code=visit_code,
                title=f"Month {index}",
                timepoint=index,
                rbase=relativedelta(months=index),
                rlower=relativedelta(days=14),
                rupper=relativedelta(days=45),
                requisitions=requisitions,
                crfs=crfs,
                crfs_missed=crfs_missed,
                facility_name="7-day-clinic",
                allow_unscheduled=True,
            )
        )
    for visit in visits:
        schedule3.add_visit(visit)

    visit_schedule3.add_schedule(schedule3)
    return visit_schedule3



from .visit_schedule1 import get_visit_schedule1
from .visit_schedule2 import get_visit_schedule2
from .visit_schedule3 import get_visit_schedule3
from .visit_schedule4 import get_visit_schedule4
from .visit_schedule5 import get_visit_schedule5
from .visit_schedule6 import get_visit_schedule6

__all__ = [
    "get_visit_schedule1",
    "get_visit_schedule2",
    "get_visit_schedule3",
    "get_visit_schedule4",
    "get_visit_schedule5",
    "get_visit_schedule6",
]



from datetime import datetime
from decimal import Context
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import FR, MO, SA, SU, TH, TU, WE, relativedelta
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase, override_settings
from edc_consent.site_consents import site_consents
from edc_facility.import_holidays import import_holidays
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED, UNSCHEDULED
from edc_visit_tracking.utils import get_related_visit_model_cls

from edc_appointment.constants import INCOMPLETE_APPT, SCHEDULED_APPT, UNSCHEDULED_APPT
from edc_appointment.exceptions import AppointmentDatetimeError
from edc_appointment.utils import get_appointment_model_cls, get_appt_reason_choices
from edc_appointment_app.consents import consent_v1
from edc_appointment_app.visit_schedule import get_visit_schedule1, get_visit_schedule2

from ..helper import Helper

utc_tz = ZoneInfo("UTC")


@override_settings(SITE_ID=10, EDC_SITES_REGISTER_DEFAULT=True)
@time_machine.travel(datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz))
class TestAppointment(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        self.subject_identifier = "12345"
        site_visit_schedules._registry = {}
        self.visit_schedule1 = get_visit_schedule1()
        self.visit_schedule2 = get_visit_schedule2()
        site_visit_schedules.register(self.visit_schedule1)
        site_visit_schedules.register(self.visit_schedule2)
        site_consents.registry = {}
        site_consents.register(consent_v1)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=ResearchProtocolConfig().study_open_datetime,
        )

    def test_appointments_dates_mo(self):
        """Test appointment datetimes are chronological."""
        for day in [MO, TU, WE, TH, FR, SA, SU]:
            helper = self.helper_cls(
                subject_identifier=f"{self.subject_identifier}-{day}",
                now=ResearchProtocolConfig().study_open_datetime,
            )
            subject_consent = helper.consent_and_put_on_schedule(
                visit_schedule_name=self.visit_schedule1.name,
                schedule_name="schedule1",
                report_datetime=get_utcnow(),
                onschedule_datetime=get_utcnow()
                + relativedelta(weeks=2)
                + relativedelta(weekday=day(-1)),
            )
            appt_datetimes = [
                obj.appt_datetime
                for obj in get_appointment_model_cls()
                .objects.filter(subject_identifier=subject_consent.subject_identifier)
                .order_by("appt_datetime")
            ]
            last = None
            self.assertGreater(len(appt_datetimes), 0)
            for appt_datetime in appt_datetimes:
                if not last:
                    last = appt_datetime
                else:
                    self.assertGreater(appt_datetime, last)
                    last = appt_datetime

    def test_attempt_to_change(self):
        for day in [MO, TU, WE, TH, FR, SA, SU]:
            helper = self.helper_cls(
                subject_identifier=f"{self.subject_identifier}-{day}",
                now=ResearchProtocolConfig().study_open_datetime,
            )
            subject_consent = helper.consent_and_put_on_schedule(
                visit_schedule_name=self.visit_schedule1.name,
                schedule_name="schedule1",
                report_datetime=get_utcnow(),
            )
        subject_identifier = subject_consent.subject_identifier
        self.assertEqual(
            get_appointment_model_cls()
            .objects.filter(subject_identifier=subject_identifier)
            .count(),
            4,
        )

        # create two unscheduled appts after first
        appointment0 = (
            get_appointment_model_cls()
            .objects.filter(subject_identifier=subject_identifier)
            .order_by("appt_datetime")[0]
        )
        get_related_visit_model_cls().objects.create(
            appointment=appointment0,
            report_datetime=appointment0.appt_datetime,
            reason=SCHEDULED,
        )
        appointment0.appt_status = INCOMPLETE_APPT
        appointment0.save()

        appointment0_1 = self.helper.add_unscheduled_appointment(appointment0)
        get_related_visit_model_cls().objects.create(
            appointment=appointment0_1,
            report_datetime=appointment0.appt_datetime,
            reason=UNSCHEDULED,
        )
        appointment0_1.appt_status = INCOMPLETE_APPT
        appointment0_1.save()

        appointment0_2 = self.helper.add_unscheduled_appointment(appointment0_1)

        appointment0_1.appt_datetime = appointment0_2.appt_datetime + relativedelta(days=1)

        self.assertRaises(AppointmentDatetimeError, appointment0_1.save)

    def test_timepoint(self):
        """Assert timepoints are saved from the schedule correctly
        as Decimals and ordered by appt_datetime.
        """
        context = Context(prec=2)
        schedule_name = self.visit_schedule1.schedules.get("schedule1").name
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule1.name, schedule_name=schedule_name
        )
        self.assertEqual(
            [
                obj.timepoint
                for obj in get_appointment_model_cls().objects.all().order_by("appt_datetime")
            ],
            [context.create_decimal(n) for n in range(0, 4)],
        )

    def test_first_appointment_with_visit_schedule(self):
        """Asserts first appointment correctly selected if just
        visit_schedule_name provided.
        """
        schedule_name = self.visit_schedule2.schedules.get("schedule2").name
        subject_consent = self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule2.name, schedule_name=schedule_name
        )

        schedule = self.visit_schedule1.schedules.get("schedule1")
        schedule.put_on_schedule(subject_consent.subject_identifier, get_utcnow())

        first_appointment = get_appointment_model_cls().objects.first_appointment(
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule2",
        )
        self.assertEqual(first_appointment.schedule_name, "schedule2")

        self.assertEqual(
            get_appointment_model_cls()
            .objects.filter(
                subject_identifier=self.subject_identifier,
                visit_schedule_name="visit_schedule2",
            )
            .order_by("appt_datetime")[0],
            first_appointment,
        )

    @override_settings(
        EDC_APPOINTMENT_APPT_REASON_CHOICES=((UNSCHEDULED_APPT, "Unscheduled"),)
    )
    def test_setting_reason_choices_invalid(self):
        self.assertRaises(ImproperlyConfigured, get_appt_reason_choices)

    @override_settings(
        EDC_APPOINTMENT_APPT_REASON_CHOICES=(
            (SCHEDULED_APPT, "Routine / Scheduled"),
            (UNSCHEDULED_APPT, "Unscheduled"),
        )
    )
    def test_setting_reason_choices_valid(self):
        try:
            choices = get_appt_reason_choices()
        except ImproperlyConfigured:
            self.fail("ImproperlyConfigured unexpectedly raised")
        self.assertEqual(
            choices,
            (
                (SCHEDULED_APPT, "Routine / Scheduled"),
                (UNSCHEDULED_APPT, "Unscheduled"),
            ),
        )

    @override_settings(
        EDC_APPOINTMENT_APPT_REASON_CHOICES=(
            (SCHEDULED_APPT, "Routine / Scheduled"),
            (UNSCHEDULED_APPT, "Unscheduled"),
            ("blah", "Blah"),
        )
    )
    def test_setting_reason_choices_valid2(self):
        try:
            choices = get_appt_reason_choices()
        except ImproperlyConfigured:
            self.fail("ImproperlyConfigured unexpectedly raised")
        self.assertEqual(
            choices,
            (
                (SCHEDULED_APPT, "Routine / Scheduled"),
                (UNSCHEDULED_APPT, "Unscheduled"),
                ("blah", "Blah"),
            ),
        )



# Generated by Django 3.2.13 on 2022-10-04 03:29

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0031_auto_20220704_1841"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="appointment",
            unique_together={
                (
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "appt_datetime",
                ),
                (
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "timepoint",
                    "visit_code_sequence",
                ),
            },
        ),
    ]



import pandas as pd
from django.apps import apps as django_apps
from django_pandas.io import read_frame
from edc_pdutils.utils import convert_dates_from_model

from ...constants import NEW_APPT
from ...utils import get_appointment_model_cls


def get_appointment_df(
    normalize: bool | None = None,
    localize: bool | None = None,
    values: list[str] | None = None,
    site_id: int | None = None,
) -> pd.DataFrame:
    normalize = True if normalize is None else normalize
    localize = True if localize is None else localize
    appointment_model_cls = get_appointment_model_cls()
    appointment_type_model_cls = django_apps.get_model("edc_appointment.appointmenttype")
    opts = {}
    if site_id:
        opts = {"site_id": site_id}
    if values:
        df_appt = read_frame(
            appointment_model_cls.objects.values(*values).filter(**opts), verbose=False
        )
    else:
        df_appt = read_frame(appointment_model_cls.objects.filter(**opts), verbose=False)
    df_appt = convert_dates_from_model(
        df_appt, appointment_model_cls, normalize=normalize, localize=localize
    )
    df_appt = df_appt.rename(
        columns={
            "id": "appointment_id",
            "site": "site_id",
        }
    )
    # rework visit code
    df_appt["visit_code_str"] = df_appt["visit_code"]
    df_appt["visit_code"] = df_appt["visit_code"].astype(float)
    df_appt["visit_code_sequence"] = df_appt["visit_code_sequence"].astype(float)
    df_appt["visit_code_sequence"] = df_appt["visit_code_sequence"].apply(
        lambda x: x / 10.0 if x > 0.0 else 0.0
    )
    df_appt["visit_code"] = df_appt["visit_code"] + df_appt["visit_code_sequence"]

    # merge in baseline_datetime
    df_baseline = df_appt.copy()
    df_baseline = df_baseline[(df_baseline["visit_code"] == 1000.0)]
    df_baseline = df_baseline.rename(columns={"appt_datetime": "baseline_datetime"})
    df_baseline = df_baseline[["subject_identifier", "baseline_datetime"]]
    df_appt = df_appt.merge(df_baseline, on="subject_identifier", how="left")

    # merge in last appointment
    df_last = (
        df_appt[df_appt.appt_status != NEW_APPT]
        .groupby("subject_identifier")
        .agg({"visit_code": "max", "appt_datetime": "max"})
    ).copy()
    df_last = df_last.rename(
        columns={"visit_code": "last_visit_code", "appt_datetime": "last_appt_datetime"}
    )
    df_last["last_visit_code_str"] = (
        df_last["last_visit_code"].astype("int64").apply(lambda x: str(x))
    )
    df_last = df_last.reset_index()
    df_appt = df_appt.merge(df_last, on="subject_identifier", how="left")

    # merge in next appointment
    df_next = (
        df_appt[df_appt.appt_status == NEW_APPT]
        .groupby("subject_identifier")
        .agg({"visit_code": "min", "appt_datetime": "min"})
    ).copy()
    df_next = df_next.rename(
        columns={"visit_code": "next_visit_code", "appt_datetime": "next_appt_datetime"}
    )
    df_next["next_visit_code_str"] = (
        df_next["next_visit_code"].astype("int64").apply(lambda x: str(x))
    )
    df_next = df_next.reset_index()
    df_appt = df_appt.merge(df_next, on="subject_identifier", how="left")

    # merge in appt type
    df1 = read_frame(appointment_type_model_cls.objects.values("id", "name").all())
    df1 = df1.rename(columns={"id": "appt_type"})
    df_appt = df_appt.merge(df1, on="appt_type", how="left")
    df_appt = df_appt.drop(columns=["appt_type"])
    df_appt = df_appt.rename(columns={"name": "appt_type"})

    return df_appt



# Generated by Django 4.2.5 on 2023-09-27 02:09

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0044_alter_appointment_appt_reason_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="appointment",
            name="appt_reason",
            field=models.CharField(
                choices=[
                    ("scheduled", "Scheduled (study-defined)"),
                    ("unscheduled", "Unscheduled / Routine"),
                ],
                help_text="The reason for visit from the visit report will be validated against this response. Refer to the protocol documentation for the definition of a scheduled appointment.",
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="comment",
            field=models.CharField(
                blank=True, max_length=250, null=True, verbose_name="Comment"
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_reason",
            field=models.CharField(
                choices=[
                    ("scheduled", "Scheduled (study-defined)"),
                    ("unscheduled", "Unscheduled / Routine"),
                ],
                help_text="The reason for visit from the visit report will be validated against this response. Refer to the protocol documentation for the definition of a scheduled appointment.",
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="comment",
            field=models.CharField(
                blank=True, max_length=250, null=True, verbose_name="Comment"
            ),
        ),
    ]



from __future__ import annotations

import datetime as dt
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings, tag
from edc_consent import site_consents
from edc_facility.import_holidays import import_holidays
from edc_metadata.models import CrfMetadata
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from edc_appointment.constants import INCOMPLETE_APPT
from edc_appointment.models import Appointment
from edc_appointment_app.consents import consent_v1
from edc_appointment_app.tests.appointment_app_test_case_mixin import (
    AppointmentAppTestCaseMixin,
)
from edc_appointment_app.visit_schedule import get_visit_schedule1

from ...creators import UnscheduledAppointmentCreator
from ...utils import reset_visit_code_sequence_or_pass
from ..helper import Helper

utc_tz = ZoneInfo("UTC")

test_datetime = dt.datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz)


@override_settings(SITE_ID=10)
@time_machine.travel(test_datetime)
class TestMoveAppointment(AppointmentAppTestCaseMixin, TestCase):
    helper_cls = Helper

    def setUp(self):
        self.subject_identifier = "12345"

        site_consents.registry = {}
        site_consents.register(consent_v1)

        site_visit_schedules._registry = {}
        self.visit_schedule = get_visit_schedule1()
        site_visit_schedules.register(self.visit_schedule)

        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=ResearchProtocolConfig().study_open_datetime,
        )

        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule.name,
            schedule_name="schedule1",
            report_datetime=get_utcnow(),
        )
        appointments = Appointment.objects.filter(subject_identifier=self.subject_identifier)
        self.assertEqual(appointments.count(), 4)

        appointment = Appointment.objects.get(timepoint=0.0)
        self.create_related_visit(appointment)
        appointment = Appointment.objects.get(timepoint=1.0)
        self.create_related_visit(appointment)

        self.appt_datetimes = [
            o.appt_datetime for o in Appointment.objects.all().order_by("appt_datetime")
        ]

    @classmethod
    def setUpTestData(cls):
        import_holidays()

    @staticmethod
    def create_unscheduled(appointment: Appointment, days: int = None):
        creator = UnscheduledAppointmentCreator(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            suggested_appt_datetime=appointment.appt_datetime + relativedelta(days=days),
            suggested_visit_code_sequence=appointment.visit_code_sequence + 1,
        )
        appointment = creator.appointment
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save_base(update_fields=["appt_status"])
        return appointment

    @staticmethod
    def get_visit_codes(by: str = None, visit_schedule_name: str | None = None, **kwargs):
        opts = dict(visit_schedule_name=visit_schedule_name)
        return [
            f"{o.visit_code}.{o.visit_code_sequence}"
            for o in Appointment.objects.filter(**opts).order_by(by)
        ]

    @tag("4")
    def test_resequence_appointment_on_insert_between_two_unscheduled(self):
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=0)
        self.assertEqual(self.create_unscheduled(appointment, days=2).visit_code_sequence, 1)
        self.assertEqual(self.create_unscheduled(appointment, days=4).visit_code_sequence, 2)
        self.assertEqual(self.create_unscheduled(appointment, days=5).visit_code_sequence, 3)

        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "1000.3", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(
                by="appt_datetime", visit_schedule_name=appointment.visit_schedule_name
            ),
        )

        # insert an appt between 1000.1 and 1000.2
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=1)
        appointment = UnscheduledAppointmentCreator(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code="1000",
            facility=appointment.facility,
            suggested_appt_datetime=appointment.appt_datetime + relativedelta(days=1),
            suggested_visit_code_sequence=2,
        )
        self.assertEqual(appointment.visit_code_sequence, 2)

        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "1000.3", "1000.4", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(
                by="appt_datetime", visit_schedule_name=appointment.visit_schedule_name
            ),
        )

    @tag("4")
    def test_repair_visit_code_sequences(self):
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=0)
        self.create_unscheduled(appointment, days=2)

        appt2 = self.create_unscheduled(appointment, days=4)
        appt3 = self.create_unscheduled(appointment, days=5)

        appt2.visit_code_sequence = 3333
        appt2.save_base(update_fields=["visit_code_sequence"])

        appt3.visit_code_sequence = 33
        appt3.save_base(update_fields=["visit_code_sequence"])

        self.assertEqual(
            ["1000.0", "1000.1", "1000.3333", "1000.33", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(
                by="appt_datetime", visit_schedule_name=appointment.visit_schedule_name
            ),
        )

        reset_visit_code_sequence_or_pass(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule.name,
            schedule_name="schedule1",
            visit_code="1000",
        )

        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "1000.3", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(
                by="appt_datetime", visit_schedule_name=appointment.visit_schedule_name
            ),
        )

    @tag("4")
    def test_repair_visit_code_sequences_with_related_visit(self):
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=0)
        appt1 = self.create_unscheduled(appointment, days=2)
        self.create_related_visit(appt1)
        appt2 = self.create_unscheduled(appointment, days=4)
        self.create_related_visit(appt2)
        appt3 = self.create_unscheduled(appointment, days=5)
        self.create_related_visit(appt3)

        appt2.visit_code_sequence = 3333
        appt2.save_base(update_fields=["visit_code_sequence"])
        appt2.related_visit.visit_code_sequence = 3333
        appt2.related_visit.save_base(update_fields=["visit_code_sequence"])

        appt3.visit_code_sequence = 33
        appt3.save_base(update_fields=["visit_code_sequence"])
        appt2.related_visit.visit_code_sequence = 33
        appt2.related_visit.save_base(update_fields=["visit_code_sequence"])

        self.assertEqual(
            ["1000.0", "1000.1", "1000.3333", "1000.33", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(
                by="appt_datetime", visit_schedule_name=appointment.visit_schedule_name
            ),
        )

        reset_visit_code_sequence_or_pass(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule.name,
            schedule_name="schedule1",
            visit_code="1000",
        )

        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "1000.3", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(
                by="appt_datetime", visit_schedule_name=appointment.visit_schedule_name
            ),
        )

        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "1000.3", "2000.0"],
            [
                f"{o.visit_code}.{o.visit_code_sequence}"
                for o in Appointment.objects.all().order_by("appt_datetime")
                if getattr(o, "related_visit", None)
            ],
        )

    @tag("4")
    def test_repair_visit_code_sequences_with_metadata(self):
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=0)
        appt1 = self.create_unscheduled(appointment, days=2)
        self.create_related_visit(appt1)
        appt2 = self.create_unscheduled(appointment, days=4)
        self.create_related_visit(appt2)
        appt3 = self.create_unscheduled(appointment, days=5)
        self.create_related_visit(appt3)

        appt2.visit_code_sequence = 3333
        appt2.save_base(update_fields=["visit_code_sequence"])
        appt2.related_visit.visit_code_sequence = 3333
        appt2.related_visit.save_base(update_fields=["visit_code_sequence"])
        CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=2).update(
            visit_code_sequence=3333
        )

        appt3.visit_code_sequence = 33
        appt3.save_base(update_fields=["visit_code_sequence"])
        appt2.related_visit.visit_code_sequence = 33
        appt2.related_visit.save_base(update_fields=["visit_code_sequence"])
        CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=3).update(
            visit_code_sequence=33
        )

        self.assertEqual(
            CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=3333).count(), 3
        )
        self.assertEqual(
            CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=33).count(), 3
        )
        reset_visit_code_sequence_or_pass(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule.name,
            schedule_name="schedule1",
            visit_code="1000",
        )
        self.assertEqual(
            CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=3333).count(), 0
        )
        self.assertEqual(
            CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=33).count(), 0
        )

        self.assertEqual(
            CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=2).count(), 3
        )
        self.assertEqual(
            CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=3).count(), 3
        )



# Generated by Django 4.1.2 on 2023-01-24 15:19

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0034_appointmenttype_alter_appointment_appt_type_and_more"),
    ]

    operations = [
        migrations.RenameField(
            model_name="appointment",
            old_name="appt_type",
            new_name="appt_type_old",
        ),
        migrations.RenameField(
            model_name="historicalappointment",
            old_name="appt_type",
            new_name="appt_type_old",
        ),
        migrations.RenameField(
            model_name="appointment",
            old_name="appt_type_new",
            new_name="appt_type",
        ),
        migrations.RenameField(
            model_name="historicalappointment",
            old_name="appt_type_new",
            new_name="appt_type",
        ),
    ]



import warnings

from .test_case_mixins import AppointmentTestCaseMixin  # noqa

warnings.warn(
    "This path/func name is deprecated in favor of test_case_mixins.AppointmentTestCaseMixin.",
    DeprecationWarning,
    stacklevel=2,
)



import datetime as dt
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings
from edc_facility import import_holidays
from edc_sites.tests import SiteTestCaseMixin
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from edc_appointment.models import Appointment
from edc_appointment.tests.helper import Helper
from edc_appointment.utils import (
    AppointmentDateWindowPeriodGapError,
    get_appointment_by_datetime,
    get_window_gap_days,
)
from edc_appointment_app.visit_schedule import get_visit_schedule4

utc = ZoneInfo("UTC")


@time_machine.travel(dt.datetime(2019, 7, 11, 8, 00, tzinfo=utc))
@override_settings(SITE_ID=10)
class TestAppointmentWindowPeriod2(SiteTestCaseMixin, TestCase):
    helper_cls = Helper

    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        self.subject_identifier = "12345"
        self.visit_schedule4 = get_visit_schedule4()
        self.schedule4 = self.visit_schedule4.schedules.get("three_monthly_schedule")
        site_visit_schedules._registry = {}
        site_visit_schedules.register(self.visit_schedule4)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=get_utcnow() - relativedelta(years=2),
        )

    @override_settings(EDC_VISIT_SCHEDULE_DEFAULT_MAX_VISIT_GAP_ALLOWED=0)
    def test_suggested_date_in_window_period_gap_raises(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule4.name,
            schedule_name=self.schedule4.name,
        )
        appointments = Appointment.objects.filter(
            subject_identifier=self.subject_identifier
        ).order_by("appt_datetime")
        self.assertEqual(appointments.count(), 5)
        appointment_1000 = appointments[0]
        appointment_1060 = appointments[2]

        subject_identifier = appointment_1000.subject_identifier
        visit_schedule_name = appointment_1000.visit_schedule_name
        schedule_name = appointment_1000.schedule_name

        suggested_appt_datetime = (
            appointment_1060.appt_datetime
            - appointment_1060.visit.rlower
            - relativedelta(days=1)
        )
        with self.assertRaises(AppointmentDateWindowPeriodGapError) as cm:
            get_appointment_by_datetime(
                suggested_appt_datetime,
                subject_identifier,
                visit_schedule_name,
                schedule_name,
            )

        self.assertIn(
            "Date falls in a `window period gap` between 1030 and 1060", str(cm.exception)
        )

    @override_settings(EDC_VISIT_SCHEDULE_DEFAULT_MAX_VISIT_GAP_ALLOWED=7)
    def test_match_window_period_gap_adjusted(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule4.name,
            schedule_name=self.schedule4.name,
        )
        appointments = Appointment.objects.filter(
            subject_identifier=self.subject_identifier
        ).order_by("appt_datetime")
        self.assertEqual(appointments.count(), 5)
        appointment_1000 = appointments[0]
        appointment_1060 = appointments[2]

        subject_identifier = appointment_1000.subject_identifier
        visit_schedule_name = appointment_1000.visit_schedule_name
        schedule_name = appointment_1000.schedule_name

        for days, in_adjusted_window in [
            (6, True),
            (7, True),
            (8, False),
            (9, False),
        ]:
            with self.subTest(days=days, in_adjusted_window=in_adjusted_window):
                suggested_appt_datetime = (
                    appointment_1060.appt_datetime
                    - appointment_1060.visit.rlower
                    - relativedelta(days=days)
                )
                try:
                    appointment = get_appointment_by_datetime(
                        suggested_appt_datetime,
                        subject_identifier,
                        visit_schedule_name,
                        schedule_name,
                        raise_if_in_gap=False,
                    )
                except AppointmentDateWindowPeriodGapError as e:
                    self.fail(
                        f"AppointmentDateWindowPeriodGapError unexpectedly raised. Got {e}"
                    )
                if not in_adjusted_window:
                    self.assertIsNone(appointment)
                else:
                    self.assertEqual("1060", appointment.visit_code)

    def test_past_last_visit(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule4.name,
            schedule_name=self.schedule4.name,
        )
        appointments = Appointment.objects.filter(
            subject_identifier=self.subject_identifier
        ).order_by("appt_datetime")
        self.assertEqual(appointments.count(), 5)
        appointment_1000 = appointments[0]
        appointment_1120 = appointments[4]

        subject_identifier = appointment_1000.subject_identifier
        visit_schedule_name = appointment_1000.visit_schedule_name
        schedule_name = appointment_1000.schedule_name

        suggested_appt_datetime = (
            appointment_1120.appt_datetime
            + appointment_1120.visit.rupper
            + relativedelta(days=1)
        )
        try:
            appointment = get_appointment_by_datetime(
                suggested_appt_datetime,
                subject_identifier,
                visit_schedule_name,
                schedule_name,
                raise_if_in_gap=False,
            )
        except AppointmentDateWindowPeriodGapError as e:
            self.fail(f"AppointmentDateWindowPeriodGapError unexpectedly raised. Got {e}")
        self.assertIsNone(appointment)

    def test_window_gap_days(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule4.name,
            schedule_name=self.schedule4.name,
        )
        appointments = Appointment.objects.filter(
            subject_identifier=self.subject_identifier
        ).order_by("appt_datetime")
        appointment_1030 = appointments[1]
        gap_days = get_window_gap_days(appointment_1030)
        self.assertEqual(gap_days, 62)



from edc_visit_schedule.tests.dummy_panel import DummyPanel
from edc_visit_schedule.visit import (
    Crf,
    CrfCollection,
    Requisition,
    RequisitionCollection,
)


class MockPanel(DummyPanel):
    """`requisition_model` is normally set when the lab profile
    is set up.
    """

    def __init__(self, name):
        super().__init__(requisition_model="edc_appointment_app.subjectrequisition", name=name)


panel_one = MockPanel(name="one")
panel_two = MockPanel(name="two")
panel_three = MockPanel(name="three")
panel_four = MockPanel(name="four")
panel_five = MockPanel(name="five")
panel_six = MockPanel(name="six")

crfs = CrfCollection(
    Crf(show_order=1, model="edc_appointment_app.crfone", required=True),
    Crf(show_order=2, model="edc_appointment_app.crftwo", required=True),
    Crf(show_order=3, model="edc_appointment_app.crfthree", required=True),
    Crf(show_order=4, model="edc_appointment_app.crffour", required=True),
    Crf(show_order=5, model="edc_appointment_app.crffive", required=True),
)

crfs_missed = CrfCollection(
    Crf(show_order=10, model="edc_appointment_app.subjectvisitmissed"),
    name="missed",
)

requisitions = RequisitionCollection(
    Requisition(show_order=10, panel=panel_one, required=True, additional=False),
    Requisition(show_order=20, panel=panel_two, required=True, additional=False),
    Requisition(show_order=30, panel=panel_three, required=True, additional=False),
    Requisition(show_order=40, panel=panel_four, required=True, additional=False),
    Requisition(show_order=50, panel=panel_five, required=True, additional=False),
    Requisition(show_order=60, panel=panel_six, required=True, additional=False),
)


crfs_unscheduled = CrfCollection(
    Crf(show_order=1, model="edc_appointment_app.crfone", required=True),
    Crf(show_order=3, model="edc_appointment_app.crfthree", required=True),
    Crf(show_order=5, model="edc_appointment_app.crffive", required=True),
)



from datetime import datetime
from decimal import Decimal
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings
from edc_consent.site_consents import site_consents
from edc_facility.import_holidays import import_holidays
from edc_sites.tests import SiteTestCaseMixin
from edc_utils import get_utcnow
from edc_visit_schedule.exceptions import ScheduleError
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED, UNSCHEDULED

from edc_appointment.constants import (
    CANCELLED_APPT,
    IN_PROGRESS_APPT,
    INCOMPLETE_APPT,
    NEW_APPT,
)
from edc_appointment.creators import UnscheduledAppointmentCreator
from edc_appointment.exceptions import (
    InvalidParentAppointmentMissingVisitError,
    InvalidParentAppointmentStatusError,
    UnscheduledAppointmentNotAllowed,
)
from edc_appointment.models import Appointment
from edc_appointment_app.consents import consent_v1
from edc_appointment_app.models import SubjectVisit
from edc_appointment_app.visit_schedule import get_visit_schedule1, get_visit_schedule2

from ..helper import Helper

utc_tz = ZoneInfo("UTC")

N = NEW_APPT
INC = INCOMPLETE_APPT


def get_appointment(visit_code: str, visit_code_sequence: int, options=None):
    return Appointment.objects.get(
        subject_identifier=options.subject_identifier,
        visit_code=visit_code,
        visit_code_sequence=visit_code_sequence,
        visit_schedule_name=options.visit_schedule1.name,
        schedule_name=options.schedule1.name,
    )


def set_to_inprogress(obj: Appointment) -> Appointment:
    obj.appt_status = IN_PROGRESS_APPT
    obj.save()
    return Appointment.objects.get(id=obj.id)


def add_subject_visit(obj: Appointment) -> Appointment:
    SubjectVisit.objects.create(
        appointment=obj,
        subject_identifier=obj.subject_identifier,
        report_datetime=obj.appt_datetime,
        reason=SCHEDULED,
        visit_code=obj.visit_code,
        visit_code_sequence=obj.visit_code_sequence,
        visit_schedule_name=obj.visit_schedule_name,
        schedule_name=obj.schedule_name,
    )
    return Appointment.objects.get(id=obj.id)


def set_to_incomplete(obj: Appointment) -> Appointment:
    # close appt (set to INCOMPLETE_APPT)
    obj.appt_status = INCOMPLETE_APPT
    obj.save()
    return Appointment.objects.get(id=obj.id)


def appt_status() -> list[str]:
    return [obj.appt_status for obj in Appointment.objects.all().order_by("appt_datetime")]


def appt_status_detailed() -> list[tuple[str, int, str]]:
    return [
        (obj.visit_code, obj.visit_code_sequence, obj.appt_status)
        for obj in Appointment.objects.all().order_by("appt_datetime")
    ]


def get_unscheduled(obj: Appointment) -> Appointment:
    creator = UnscheduledAppointmentCreator(
        subject_identifier=obj.subject_identifier,
        visit_schedule_name=obj.visit_schedule_name,
        schedule_name=obj.schedule_name,
        visit_code=obj.visit_code,
        facility=obj.facility,
        suggested_visit_code_sequence=1,
    )
    return creator.appointment


@time_machine.travel(datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz))
@override_settings(SITE_ID=10)
class TestUnscheduledAppointmentCreator(SiteTestCaseMixin, TestCase):
    helper_cls = Helper

    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        self.subject_identifier = "12345"
        self.visit_schedule1 = get_visit_schedule1()
        self.schedule1 = self.visit_schedule1.schedules.get("schedule1")
        self.visit_schedule2 = get_visit_schedule2()
        self.schedule2 = self.visit_schedule2.schedules.get("schedule2")
        site_consents.registry = {}
        site_consents.register(consent_v1)
        site_visit_schedules._registry = {}
        site_visit_schedules.register(self.visit_schedule1)
        site_visit_schedules.register(self.visit_schedule2)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=datetime(2017, 1, 7, tzinfo=ZoneInfo("UTC")),
        )

    def test_unscheduled_allowed_but_raises_on_appt_status(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule1.name,
            schedule_name=self.schedule1.name,
        )
        visit = self.visit_schedule1.schedules.get(self.schedule1.name).visits.first
        appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code=visit.code,
            visit_code_sequence=0,
        )
        # subject_visit not created so expect exception because of
        # the missing subject_visit
        for appt_status in [NEW_APPT, IN_PROGRESS_APPT, CANCELLED_APPT]:
            with self.subTest(appt_status=appt_status):
                appointment.appt_status = appt_status
                appointment.save()
                self.assertEqual(appointment.appt_status, appt_status)
                self.assertRaises(
                    InvalidParentAppointmentMissingVisitError,
                    UnscheduledAppointmentCreator,
                    subject_identifier=self.subject_identifier,
                    visit_schedule_name=self.visit_schedule1.name,
                    schedule_name=self.schedule1.name,
                    visit_code=visit.code,
                    suggested_visit_code_sequence=appointment.visit_code_sequence + 1,
                )
        # add a subject_visit and expect exception to be raises because
        # of appt_status
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        appointment.refresh_from_db()
        self.assertEqual(appointment.related_visit, subject_visit)
        for appt_status in [NEW_APPT, INCOMPLETE_APPT, IN_PROGRESS_APPT, CANCELLED_APPT]:
            with self.subTest(appt_status=appt_status):
                appointment.appt_status = appt_status
                appointment.save()
                if appointment.appt_status == INCOMPLETE_APPT:
                    continue
                self.assertEqual(appointment.appt_status, appt_status)
                self.assertRaises(
                    InvalidParentAppointmentStatusError,
                    UnscheduledAppointmentCreator,
                    subject_identifier=self.subject_identifier,
                    visit_schedule_name=self.visit_schedule1.name,
                    schedule_name=self.schedule1.name,
                    visit_code=visit.code,
                    suggested_visit_code_sequence=appointment.visit_code_sequence + 1,
                )

    def test_unscheduled_not_allowed(self):
        self.assertRaises(
            UnscheduledAppointmentNotAllowed,
            UnscheduledAppointmentCreator,
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule2.name,
            schedule_name=self.schedule2.name,
            visit_code="5000",
            suggested_visit_code_sequence=1,
        )

    def test_appt_status(self):
        start_date = datetime(2017, 1, 7, tzinfo=ZoneInfo("UTC"))
        traveller = time_machine.travel(start_date)
        traveller.start()
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule1.name, schedule_name=self.schedule1.name
        )
        traveller.stop()
        for visit in self.visit_schedule1.schedules.get(self.schedule1.name).visits.values():
            # with self.subTest(visit=visit):
            # get parent appointment
            new_appointment = None
            appointment = Appointment.objects.get(
                subject_identifier=self.subject_identifier,
                visit_code=visit.code,
                visit_code_sequence=0,
                timepoint=visit.timepoint,
                visit_schedule_name=self.visit_schedule1.name,
                schedule_name=self.schedule1.name,
            )
            traveller = time_machine.travel(appointment.appt_datetime)
            traveller.start()
            appointment.appt_status = IN_PROGRESS_APPT
            appointment.save()
            appointment.refresh_from_db()
            # fill in subject visit report for this appointment
            subject_visit = SubjectVisit.objects.create(
                appointment=appointment,
                subject_identifier=self.subject_identifier,
                report_datetime=appointment.appt_datetime,
                reason=SCHEDULED,
                visit_code=visit.code,
                visit_code_sequence=0,
                visit_schedule_name=self.visit_schedule1.name,
                schedule_name=self.schedule1.name,
            )
            appointment.refresh_from_db()
            self.assertTrue(appointment.related_visit, subject_visit)
            self.assertEqual(0, appointment.related_visit.visit_code_sequence)
            self.assertEqual(1, appointment.next_visit_code_sequence)

            # close appt (set to INCOMPLETE_APPT)
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save()
            appointment.refresh_from_db()

            print(visit, appointment, appointment.related_visit, appointment.appt_status)

            traveller.stop()
            traveller = time_machine.travel(appointment.appt_datetime + relativedelta(days=1))
            traveller.start()

            # create unscheduled off of this appt
            creator = UnscheduledAppointmentCreator(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule1.name,
                schedule_name=self.schedule1.name,
                visit_code=visit.code,
                facility=appointment.facility,
                suggested_visit_code_sequence=1,
            )
            new_appointment = creator.appointment
            new_appointment.appt_status = IN_PROGRESS_APPT
            new_appointment.save()
            new_appointment.refresh_from_db()
            print(
                visit,
                new_appointment,
                new_appointment.related_visit,
                new_appointment.appt_status,
            )
            if new_appointment.appt_status != IN_PROGRESS_APPT:
                new_appointment.appt_status = IN_PROGRESS_APPT
                new_appointment.save()
                new_appointment.refresh_from_db()

            self.assertEqual(new_appointment.appt_status, IN_PROGRESS_APPT)

            # submit subject visit for the unscheduled appt
            subject_visit = SubjectVisit.objects.create(
                appointment=new_appointment,
                report_datetime=get_utcnow(),
                reason=UNSCHEDULED,
                visit_code=new_appointment.visit_code,
                visit_code_sequence=new_appointment.visit_code_sequence,
                visit_schedule_name=new_appointment.visit_schedule_name,
                schedule_name=new_appointment.schedule_name,
            )
            self.assertEqual(1, new_appointment.visit_code_sequence)
            self.assertEqual(1, subject_visit.visit_code_sequence)

            # close the unscheduled appt (set to INCOMPLETE_APPT)
            new_appointment.appt_status = INCOMPLETE_APPT
            new_appointment.save()
            new_appointment.refresh_from_db()
            self.assertEqual(new_appointment.appt_status, INCOMPLETE_APPT)
            self.assertEqual(visit.timepoint, int(new_appointment.timepoint))
            traveller.stop()

    def test_appt_status2(self):

        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule1.name, schedule_name=self.schedule1.name
        )

        # 1000
        self.assertEqual(appt_status(), [N, N, N, N])
        appointment = get_appointment("1000", 0, options=self)
        appointment = set_to_inprogress(appointment)
        self.assertEqual(appt_status(), [IN_PROGRESS_APPT, N, N, N])
        appointment = add_subject_visit(appointment)
        self.assertEqual(appt_status(), [IN_PROGRESS_APPT, N, N, N])
        appointment = set_to_incomplete(appointment)
        self.assertEqual(appt_status(), [INC, N, N, N])
        unscheduled_appointment = get_unscheduled(appointment)
        self.assertEqual(appt_status(), [INC, N, N, N, N])
        unscheduled_appointment = set_to_inprogress(unscheduled_appointment)
        self.assertEqual(appt_status(), [INC, IN_PROGRESS_APPT, N, N, N])
        unscheduled_appointment = add_subject_visit(unscheduled_appointment)
        self.assertEqual(appt_status(), [INC, IN_PROGRESS_APPT, N, N, N])
        set_to_incomplete(unscheduled_appointment)
        self.assertEqual(appt_status(), [INC, INC, N, N, N])

    def test_appt_status3(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule1.name, schedule_name=self.schedule1.name
        )
        # 1000
        appointment = get_appointment("1000", 0, options=self)
        appointment = set_to_inprogress(appointment)
        appointment = add_subject_visit(appointment)
        appointment = set_to_incomplete(appointment)
        unscheduled_appointment = get_unscheduled(appointment)
        unscheduled_appointment = set_to_inprogress(unscheduled_appointment)
        unscheduled_appointment = add_subject_visit(unscheduled_appointment)
        set_to_incomplete(unscheduled_appointment)

        # 2000
        self.assertEqual(appt_status(), [INC, INC, N, N, N])
        appointment = get_appointment("2000", 0, options=self)
        appointment = set_to_inprogress(appointment)
        self.assertEqual(appt_status(), [INC, INC, IN_PROGRESS_APPT, N, N])
        appointment = add_subject_visit(appointment)
        self.assertEqual(appt_status(), [INC, INC, IN_PROGRESS_APPT, N, N])
        appointment = set_to_incomplete(appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, N, N])
        unscheduled_appointment = get_unscheduled(appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, N, N, N])
        unscheduled_appointment = set_to_inprogress(unscheduled_appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, IN_PROGRESS_APPT, N, N])
        unscheduled_appointment = add_subject_visit(unscheduled_appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, IN_PROGRESS_APPT, N, N])
        set_to_incomplete(unscheduled_appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, N, N])

    def test_appt_status4(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule1.name, schedule_name=self.schedule1.name
        )
        for visit_code in ["1000", "2000"]:
            appointment = get_appointment(visit_code, 0, options=self)
            appointment = set_to_inprogress(appointment)
            appointment = add_subject_visit(appointment)
            appointment = set_to_incomplete(appointment)
            unscheduled_appointment = get_unscheduled(appointment)
            unscheduled_appointment = set_to_inprogress(unscheduled_appointment)
            unscheduled_appointment = add_subject_visit(unscheduled_appointment)
            set_to_incomplete(unscheduled_appointment)

        # 3000
        self.assertEqual(appt_status(), [INC, INC, INC, INC, N, N])
        appointment = get_appointment("3000", 0, options=self)
        appointment = set_to_inprogress(appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, IN_PROGRESS_APPT, N])
        appointment = add_subject_visit(appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, IN_PROGRESS_APPT, N])
        appointment = set_to_incomplete(appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, INC, N])
        unscheduled_appointment = get_unscheduled(appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, INC, N, N])
        unscheduled_appointment = set_to_inprogress(unscheduled_appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, INC, IN_PROGRESS_APPT, N])
        unscheduled_appointment = add_subject_visit(unscheduled_appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, INC, IN_PROGRESS_APPT, N])
        set_to_incomplete(unscheduled_appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, INC, INC, N])

    def test_appt_status5(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule1.name, schedule_name=self.schedule1.name
        )
        for visit_code in ["1000", "2000", "3000"]:
            appointment = get_appointment(visit_code, 0, options=self)
            appointment = set_to_inprogress(appointment)
            appointment = add_subject_visit(appointment)
            appointment = set_to_incomplete(appointment)
            unscheduled_appointment = get_unscheduled(appointment)
            unscheduled_appointment = set_to_inprogress(unscheduled_appointment)
            unscheduled_appointment = add_subject_visit(unscheduled_appointment)
            set_to_incomplete(unscheduled_appointment)

        # 4000
        self.assertEqual(appt_status(), [INC, INC, INC, INC, INC, INC, N])
        appointment = get_appointment("4000", 0, options=self)
        appointment = set_to_inprogress(appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, INC, INC, IN_PROGRESS_APPT])
        appointment = add_subject_visit(appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, INC, INC, IN_PROGRESS_APPT])
        appointment = set_to_incomplete(appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, INC, INC, INC])
        unscheduled_appointment = get_unscheduled(appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, INC, INC, INC, N])
        unscheduled_appointment = set_to_inprogress(unscheduled_appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, INC, INC, INC, IN_PROGRESS_APPT])
        unscheduled_appointment = add_subject_visit(unscheduled_appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, INC, INC, INC, IN_PROGRESS_APPT])
        set_to_incomplete(unscheduled_appointment)
        self.assertEqual(appt_status(), [INC, INC, INC, INC, INC, INC, INC, INC])

    def test_unscheduled_timepoint_not_incremented(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule1.name, schedule_name=self.schedule1.name
        )
        visit = self.visit_schedule1.schedules.get(self.schedule1.name).visits.first
        appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code=visit.code
        )
        self.assertEqual(appointment.timepoint, Decimal("0.0"))
        SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()
        for index in range(1, 5):
            with self.subTest(index=index):
                creator = UnscheduledAppointmentCreator(
                    subject_identifier=appointment.subject_identifier,
                    visit_schedule_name=appointment.visit_schedule_name,
                    schedule_name=appointment.schedule_name,
                    visit_code=appointment.visit_code,
                    suggested_visit_code_sequence=index,
                    facility=appointment.facility,
                )
                self.assertEqual(appointment.timepoint, creator.appointment.timepoint)
                self.assertNotEqual(
                    appointment.visit_code_sequence,
                    creator.appointment.visit_code_sequence,
                )
                self.assertEqual(
                    creator.appointment.visit_code_sequence,
                    appointment.visit_code_sequence + 1,
                )
                SubjectVisit.objects.create(
                    appointment=creator.appointment,
                    report_datetime=get_utcnow(),
                    reason=UNSCHEDULED,
                )
                creator.appointment.appt_status = INCOMPLETE_APPT
                creator.appointment.save()
                appointment = creator.appointment

    def test_appointment_title(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule1.name, schedule_name=self.schedule1.name
        )
        appointment = Appointment.objects.first_appointment(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule1.name,
            schedule_name=self.schedule1.name,
        )
        self.assertEqual(appointment.title, "Day 1")

        SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()

        creator = UnscheduledAppointmentCreator(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            facility=appointment.facility,
            suggested_visit_code_sequence=appointment.visit_code_sequence + 1,
        )
        self.assertEqual(creator.appointment.title, "Day 1.1")

        SubjectVisit.objects.create(
            appointment=creator.appointment, report_datetime=get_utcnow(), reason=UNSCHEDULED
        )
        creator.appointment.appt_status = INCOMPLETE_APPT
        creator.appointment.save()

        next_appointment = Appointment.objects.next_appointment(
            visit_code=appointment.visit_code,
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule1.name,
            schedule_name=self.schedule1.name,
        )

        SubjectVisit.objects.create(
            appointment=next_appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        next_appointment.appt_status = INCOMPLETE_APPT
        next_appointment.save()

        creator = UnscheduledAppointmentCreator(
            subject_identifier=next_appointment.subject_identifier,
            visit_schedule_name=next_appointment.visit_schedule_name,
            schedule_name=next_appointment.schedule_name,
            visit_code=next_appointment.visit_code,
            facility=next_appointment.facility,
            suggested_visit_code_sequence=next_appointment.visit_code_sequence + 1,
        )

        self.assertEqual(creator.appointment.title, "Day 2.1")

    def test_appointment_title_if_visit_schedule_changes(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=self.visit_schedule1.name, schedule_name=self.schedule1.name
        )
        appointment = Appointment.objects.first_appointment(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule1.name,
            schedule_name=self.schedule1.name,
        )
        self.assertEqual(appointment.title, "Day 1")

        SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()

        next_appointment = Appointment.objects.next_appointment(
            visit_code=appointment.visit_code,
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule1.name,
            schedule_name=self.schedule1.name,
        )

        SubjectVisit.objects.create(
            appointment=next_appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        next_appointment.appt_status = INCOMPLETE_APPT
        next_appointment.visit_code = "1111"
        self.assertRaises(ScheduleError, next_appointment.save)



import sys
from datetime import datetime
from pathlib import Path
from zoneinfo import ZoneInfo

from edc_test_settings.default_test_settings import DefaultTestSettings

utc = ZoneInfo("UTC")

app_name = "edc_appointment"
base_dir = Path(__file__).absolute().parent

project_settings = DefaultTestSettings(
    calling_file=__file__,
    DEBUG=True,
    BASE_DIR=base_dir,
    APP_NAME=app_name,
    ETC_DIR=base_dir / "etc",
    DJANGO_CRYPTO_FIELDS_KEY_PATH=base_dir / "etc",
    GIT_DIR=base_dir.parent.parent,
    HOLIDAY_FILE=base_dir / "holidays.csv",
    EDC_RANDOMIZATION_LIST_PATH=base_dir / "etc",
    SILENCED_SYSTEM_CHECKS=[
        "sites.E101",
        "edc_sites.E001",
        "edc_navbar.E002",
        "edc_navbar.E003",
    ],
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=datetime(2016, 10, 2, 0, 0, 0, tzinfo=utc),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=datetime(2023, 10, 2, 0, 0, 0, tzinfo=utc),
    EDC_AUTH_SKIP_SITE_AUTHS=True,
    EDC_AUTH_SKIP_AUTH_UPDATER=True,
    SUBJECT_SCREENING_MODEL="edc_appointment_app.subjectscreening",
    SUBJECT_CONSENT_MODEL="edc_appointment_app.subjectconsent",
    SUBJECT_VISIT_MODEL="edc_appointment_app.subjectvisit",
    SUBJECT_VISIT_MISSED_MODEL="edc_appointment_app.subjectvisitmissed",
    SUBJECT_REQUISITION_MODEL="edc_appointment_app.subjectrequisition",
    SUBJECT_APP_LABEL="edc_appointment_app",
    INSTALLED_APPS=[
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
        "django.contrib.sites",
        "django_crypto_fields.apps.AppConfig",
        "django_revision.apps.AppConfig",
        "django_pylabels.apps.AppConfig",
        "multisite",
        "edc_sites.apps.AppConfig",
        "edc_pylabels.apps.AppConfig",
        "edc_auth.apps.AppConfig",
        "edc_action_item.apps.AppConfig",
        "edc_adverse_event.apps.AppConfig",
        "adverse_event_app.apps.AppConfig",
        "edc_offstudy.apps.AppConfig",
        "edc_consent.apps.AppConfig",
        "edc_crf.apps.AppConfig",
        "edc_dashboard.apps.AppConfig",
        "edc_data_manager.apps.AppConfig",
        "edc_lab.apps.AppConfig",
        "edc_subject_dashboard.apps.AppConfig",
        "edc_device.apps.AppConfig",
        "edc_facility.apps.AppConfig",
        "edc_form_runners.apps.AppConfig",
        "edc_identifier.apps.AppConfig",
        "edc_list_data.apps.AppConfig",
        "edc_listboard.apps.AppConfig",
        "edc_locator.apps.AppConfig",
        "edc_metadata.apps.AppConfig",
        "edc_model_admin.apps.AppConfig",
        "edc_navbar.apps.AppConfig",
        "edc_protocol.apps.AppConfig",
        "edc_randomization.apps.AppConfig",
        "edc_registration.apps.AppConfig",
        "edc_notification.apps.AppConfig",
        "edc_review_dashboard.apps.AppConfig",
        "edc_timepoint.apps.AppConfig",
        "edc_visit_schedule.apps.AppConfig",
        "edc_visit_tracking.apps.AppConfig",
        "edc_pharmacy.apps.AppConfig",
        "edc_appointment.apps.AppConfig",
        "edc_appointment_app.apps.AppConfig",
        "edc_appconfig.apps.AppConfig",
    ],
    add_dashboard_middleware=True,
    use_test_urls=True,
).settings

for k, v in project_settings.items():
    setattr(sys.modules[__name__], k, v)



import datetime as dt
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.contrib.auth.models import User
from django.test import TestCase, override_settings, tag
from edc_consent.consent_definition import ConsentDefinition
from edc_consent.consent_definition_extension import ConsentDefinitionExtension
from edc_consent.site_consents import site_consents
from edc_constants.constants import FEMALE, MALE, YES
from edc_facility import import_holidays
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils import get_utcnow
from edc_visit_schedule.post_migrate_signals import populate_visit_schedule
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED
from edc_visit_tracking.utils import get_related_visit_model_cls

from edc_appointment.models import Appointment
from edc_appointment.utils import refresh_appointments
from edc_appointment_app.models import SubjectConsentV1, SubjectConsentV1Ext
from edc_appointment_app.visit_schedule import get_visit_schedule6

utc = ZoneInfo("UTC")
tz = ZoneInfo("Africa/Dar_es_Salaam")


@override_settings(SITE_ID=10)
class TestNextAppointmentCrf(TestCase):
    @classmethod
    def setUpTestData(cls):
        import_holidays()

    @time_machine.travel(dt.datetime(2019, 6, 11, 8, 00, tzinfo=utc))
    def setUp(self):
        self.user = User.objects.create_superuser("user_login", "u@example.com", "pass")

        self.consent_v1 = ConsentDefinition(
            "edc_appointment_app.subjectconsentv1",
            version="1",
            start=ResearchProtocolConfig().study_open_datetime,
            end=ResearchProtocolConfig().study_close_datetime,
            age_min=18,
            age_is_adult=18,
            age_max=64,
            gender=[MALE, FEMALE],
        )

        consent_v1_ext = ConsentDefinitionExtension(
            "edc_appointment_app.subjectconsentv1ext",
            version="1.1",
            start=self.consent_v1.start + relativedelta(months=2),
            extends=self.consent_v1,
            timepoints=[4],
        )
        site_consents.registry = {}
        site_consents.register(self.consent_v1, extended_by=consent_v1_ext)

        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False
        site_visit_schedules.register(get_visit_schedule6(self.consent_v1))
        populate_visit_schedule()

        self.subject_identifier = "101-40990029-4"
        identity = "123456789"
        self.subject_consent = SubjectConsentV1.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow() - relativedelta(days=10),
            identity=identity,
            confirm_identity=identity,
            dob=get_utcnow() - relativedelta(years=25),
        )

        # put subject on schedule
        _, schedule = site_visit_schedules.get_by_onschedule_model(
            "edc_appointment_app.onschedulesix"
        )
        schedule.put_on_schedule(
            subject_identifier=self.subject_consent.subject_identifier,
            onschedule_datetime=self.subject_consent.consent_datetime,
        )

    @tag("1")
    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.nextappointmentcrf": ("appt_date", "visitschedule")
        }
    )
    @time_machine.travel(dt.datetime(2019, 6, 11, 8, 00, tzinfo=utc))
    def test_ok(self):
        self.assertEqual(4, Appointment.objects.all().count())
        subject_visit_model_cls = get_related_visit_model_cls()

        appointment = Appointment.objects.get(timepoint=0)
        subject_visit_model_cls.objects.create(
            report_datetime=appointment.appt_datetime,
            appointment=appointment,
            reason=SCHEDULED,
        )
        appointment = Appointment.objects.get(timepoint=1)
        subject_visit_model_cls.objects.create(
            report_datetime=appointment.appt_datetime,
            appointment=appointment,
            reason=SCHEDULED,
        )
        appointment = Appointment.objects.get(timepoint=2)
        subject_visit_model_cls.objects.create(
            report_datetime=appointment.appt_datetime,
            appointment=appointment,
            reason=SCHEDULED,
        )

        traveller = time_machine.travel(appointment.appt_datetime + relativedelta(days=10))
        traveller.start()
        SubjectConsentV1Ext.objects.create(
            subject_consent=self.subject_consent,
            report_datetime=get_utcnow(),
            site_id=self.subject_consent.site_id,
            agrees_to_extension=YES,
        )
        refresh_appointments(
            subject_identifier=self.subject_consent.subject_identifier,
            visit_schedule_name="visit_schedule6",
            schedule_name="schedule6",
        )

        self.assertEqual(5, Appointment.objects.all().count())
        traveller.stop()

        appointment = Appointment.objects.get(timepoint=3)
        traveller = time_machine.travel(appointment.appt_datetime)
        traveller.start()
        subject_visit = subject_visit_model_cls.objects.create(
            report_datetime=get_utcnow(),
            appointment=appointment,
            reason=SCHEDULED,
        )
        self.assertEqual(subject_visit.consent_version, "1.1")



from __future__ import annotations

from typing import TYPE_CHECKING

from dateutil.relativedelta import relativedelta
from edc_visit_schedule.constants import MONTH0, MONTH1, MONTH2, MONTH3, MONTH4
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.visit import Crf, CrfCollection, Visit
from edc_visit_schedule.visit_schedule import VisitSchedule

from edc_appointment_app.consents import consent_v1

if TYPE_CHECKING:
    from edc_consent.consent_definition import ConsentDefinition


def get_visit_schedule6(cdef: ConsentDefinition | None = None) -> VisitSchedule:
    app_label = "edc_appointment_app"

    cdef = cdef or consent_v1

    crfs = CrfCollection(
        Crf(show_order=1, model=f"{app_label}.nextappointmentcrf", required=True),
    )

    visit0 = Visit(
        code=MONTH0,
        title=MONTH0,
        timepoint=0,
        rbase=relativedelta(days=0),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        crfs=crfs,
        facility_name="5-day-clinic",
    )

    visit1 = Visit(
        code=MONTH1,
        title=MONTH1,
        timepoint=1,
        rbase=relativedelta(months=1),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        crfs=crfs,
        facility_name="5-day-clinic",
    )

    visit2 = Visit(
        code=MONTH2,
        title=MONTH2,
        timepoint=2,
        rbase=relativedelta(months=2),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        crfs=crfs,
        facility_name="5-day-clinic",
    )

    visit3 = Visit(
        code=MONTH3,
        title=MONTH3,
        timepoint=3,
        rbase=relativedelta(months=3),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        crfs=crfs,
        facility_name="5-day-clinic",
    )

    visit4 = Visit(
        code=MONTH4,
        title=MONTH4,
        timepoint=4,
        rbase=relativedelta(months=4),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        crfs=crfs,
        facility_name="5-day-clinic",
    )

    schedule = Schedule(
        name="schedule6",
        onschedule_model=f"{app_label}.onschedulesix",
        offschedule_model=f"{app_label}.offschedulesix",
        consent_definitions=[cdef],
        appointment_model="edc_appointment.appointment",
    )

    schedule.add_visit(visit0)
    schedule.add_visit(visit1)
    schedule.add_visit(visit2)
    schedule.add_visit(visit3)
    schedule.add_visit(visit4)

    visit_schedule6 = VisitSchedule(
        name="visit_schedule6",
        offstudy_model=f"{app_label}.subjectoffstudy",
        death_report_model=f"{app_label}.deathreport",
    )

    visit_schedule6.add_schedule(schedule)
    return visit_schedule6



from __future__ import annotations

from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING, Any

from edc_visit_tracking.constants import UNSCHEDULED

from ..constants import IN_PROGRESS_APPT, INCOMPLETE_APPT
from ..creators import UnscheduledAppointmentCreator
from ..utils import get_appointment_model_cls

if TYPE_CHECKING:
    from edc_appointment.models import Appointment


class AppointmentTestCaseMixin:
    def get_timepoint_from_visit_code(
        self: Any,
        instance: Any,
        visit_code: str,
    ) -> float | Decimal | None:
        timepoint = None
        for v in instance.schedule.visits.timepoints:
            if v.name == visit_code:
                timepoint = v.timepoint
                break
        return timepoint

    def get_appointment(
        self,
        subject_identifier: str | None = None,
        visit_code: str | None = None,
        visit_code_sequence: int | None = None,
        reason: str | None = None,
        appt_datetime: datetime | None = None,
        timepoint: float | Decimal | None = None,
    ) -> Appointment:
        if timepoint is not None:
            appointment = get_appointment_model_cls().objects.get(
                subject_identifier=subject_identifier,
                timepoint=timepoint,
                visit_code_sequence=visit_code_sequence,
            )
        else:
            appointment = get_appointment_model_cls().objects.get(
                subject_identifier=subject_identifier,
                visit_code=visit_code,
                visit_code_sequence=visit_code_sequence,
            )
        if appt_datetime:
            appointment.appt_datetime = appt_datetime
            appointment.save()
            appointment.refresh_from_db()
        if reason == UNSCHEDULED:
            appointment = self.create_unscheduled_appointment(appointment)
        appointment.appt_status = IN_PROGRESS_APPT
        appointment.save()
        appointment.refresh_from_db()
        return appointment

    @staticmethod
    def create_unscheduled_appointment(appointment: Appointment) -> Appointment:
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()
        appointment.refresh_from_db()
        appt_creator = UnscheduledAppointmentCreator(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            suggested_visit_code_sequence=appointment.visit_code_sequence + 1,
            facility=appointment.facility,
        )
        return appt_creator.appointment



from dateutil.relativedelta import relativedelta
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.visit import Visit
from edc_visit_schedule.visit_schedule import VisitSchedule

from ..consents import consent_v1
from .crfs import crfs, crfs_missed, requisitions


def get_visit_schedule2() -> VisitSchedule:
    visit_schedule2 = VisitSchedule(
        name="visit_schedule2",
        offstudy_model="edc_appointment_app.subjectoffstudy",
        death_report_model="edc_appointment_app.deathreport",
    )

    schedule2 = Schedule(
        name="schedule2",
        onschedule_model="edc_appointment_app.onscheduletwo",
        offschedule_model="edc_appointment_app.offscheduletwo",
        appointment_model="edc_appointment.appointment",
        consent_definitions=[consent_v1],
        base_timepoint=4,
    )

    visits = []
    for index in range(4, 8):
        visits.append(
            Visit(
                code=f"{1 if index == 0 else index + 1}000",
                title=f"Day {1 if index == 0 else index + 1}",
                timepoint=index,
                rbase=relativedelta(days=7 * index),
                rlower=relativedelta(days=0),
                rupper=relativedelta(days=6),
                requisitions=requisitions,
                crfs=crfs,
                crfs_missed=crfs_missed,
                facility_name="7-day-clinic",
            )
        )
    for visit in visits:
        schedule2.add_visit(visit)

    visit_schedule2.add_schedule(schedule2)
    return visit_schedule2



# Generated by Django 4.1.2 on 2022-10-07 03:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0032_alter_appointment_unique_together"),
    ]

    operations = [
        migrations.AlterField(
            model_name="appointment",
            name="appt_reason",
            field=models.CharField(
                choices=[("scheduled", "Routine / Scheduled"), ("unscheduled", "Unscheduled")],
                help_text="The reason for visit from the visit report will be validated against this response. Refer to the protocol documentation for the definition of a scheduled appointment.",
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="appt_timing",
            field=models.CharField(
                choices=[("ontime", "On time (within window period)"), ("missed", "Missed")],
                default="ontime",
                help_text="If late, you may be required to complete a protocol incident report. Refer to the protocol documentation for the allowed window periods of scheduled appointments.",
                max_length=25,
                verbose_name="Timing",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_reason",
            field=models.CharField(
                choices=[("scheduled", "Routine / Scheduled"), ("unscheduled", "Unscheduled")],
                help_text="The reason for visit from the visit report will be validated against this response. Refer to the protocol documentation for the definition of a scheduled appointment.",
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_timing",
            field=models.CharField(
                choices=[("ontime", "On time (within window period)"), ("missed", "Missed")],
                default="ontime",
                help_text="If late, you may be required to complete a protocol incident report. Refer to the protocol documentation for the allowed window periods of scheduled appointments.",
                max_length=25,
                verbose_name="Timing",
            ),
        ),
    ]



from datetime import date

from django.db import models
from django.db.models import Manager
from django.db.models.deletion import PROTECT
from edc_consent.managers import ConsentObjectsByCdefManager, CurrentSiteByCdefManager
from edc_consent.model_mixins import (
    ConsentExtensionModelMixin,
    RequiresConsentFieldsModelMixin,
)
from edc_crf.model_mixins import CrfModelMixin
from edc_identifier.managers import SubjectIdentifierManager
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_list_data.model_mixins import ListModelMixin
from edc_metadata.model_mixins.creates import CreatesMetadataModelMixin
from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_offstudy.model_mixins import OffstudyModelMixin
from edc_registration.model_mixins import UpdatesOrCreatesRegistrationModelMixin
from edc_screening.model_mixins import ScreeningModelMixin
from edc_sites.managers import CurrentSiteManager
from edc_sites.model_mixins import SiteModelMixin
from edc_utils import get_utcnow
from edc_visit_schedule.model_mixins import (
    OffScheduleModelMixin,
    OnScheduleModelMixin,
    VisitCodeFieldsModelMixin,
)
from edc_visit_schedule.models import VisitSchedule
from edc_visit_tracking.model_mixins import (
    SubjectVisitMissedModelMixin,
    VisitModelMixin,
)
from edc_visit_tracking.models import SubjectVisitMissedReasons

from edc_appointment.model_mixins.next_appointment_crf_model_mixin import (
    NextAppointmentCrfModelMixin,
)
from edc_appointment.utils import get_appointment_model_name

from .consents import consent_v1


class Panel(ListModelMixin):
    class Meta:
        pass


#
class SubjectVisit(
    SiteModelMixin,
    VisitModelMixin,
    CreatesMetadataModelMixin,
    RequiresConsentFieldsModelMixin,
    BaseUuidModel,
):
    appointment = models.OneToOneField(get_appointment_model_name(), on_delete=PROTECT)

    subject_identifier = models.CharField(max_length=25, null=True)

    report_datetime = models.DateTimeField()

    reason = models.CharField(max_length=25, null=True)


# leave this model so ensure tests pass even though a proxy related
# visit exists (but is not used)
class SubjectVisit2(SubjectVisit):
    class Meta:
        proxy = True


class SubjectRequisition(
    SiteModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    VisitCodeFieldsModelMixin,
    BaseUuidModel,
):
    @classmethod
    def related_visit_model_attr(cls):
        return "subject_visit"

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=models.PROTECT, related_name="+")

    panel = models.ForeignKey(Panel, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    is_drawn = models.CharField(max_length=15, default="YES")

    class Meta(BaseUuidModel.Meta, NonUniqueSubjectIdentifierFieldMixin.Meta):
        pass


class SubjectVisitMissed(SiteModelMixin, SubjectVisitMissedModelMixin, BaseUuidModel):
    subject_visit = models.OneToOneField(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    missed_reasons = models.ManyToManyField(
        SubjectVisitMissedReasons, blank=True, related_name="missed_reasons"
    )

    @classmethod
    def related_visit_model_attr(cls):
        return "subject_visit"

    @property
    def related_visit(self):
        return self.subject_visit

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Missed Visit Report"
        verbose_name_plural = "Missed Visit Report"


class SubjectConsent(
    SiteModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    BaseUuidModel,
):
    report_datetime = models.DateTimeField(default=get_utcnow)

    consent_datetime = models.DateTimeField(default=get_utcnow)

    version = models.CharField(max_length=25, default="1")

    identity = models.CharField(max_length=25)

    confirm_identity = models.CharField(max_length=25)

    dob = models.DateField(default=date(1995, 1, 1))


class SubjectConsentV1(SubjectConsent):

    objects = ConsentObjectsByCdefManager()
    on_site = CurrentSiteByCdefManager()

    class Meta:
        proxy = True


class SubjectConsentV1Ext(ConsentExtensionModelMixin, SiteModelMixin, BaseUuidModel):

    subject_consent = models.ForeignKey(SubjectConsentV1, on_delete=models.PROTECT)

    on_site = CurrentSiteManager()
    history = HistoricalRecords()
    objects = Manager()

    class Meta(ConsentExtensionModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = "Subject Consent Extension V1.1"
        verbose_name_plural = "Subject Consent Extension V1.1"


class SubjectScreening(ScreeningModelMixin, BaseUuidModel):
    consent_definition = consent_v1
    objects = SubjectIdentifierManager()


class SubjectOffstudy(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class SubjectOffstudy2(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class SubjectOffstudyFive(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class SubjectOffstudySix(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class SubjectOffstudySeven(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class DeathReport(SiteModelMixin, BaseUuidModel):
    subject_identifier = models.CharField(max_length=25, null=True)

    report_datetime = models.DateTimeField()


# visit_schedule


class OnSchedule(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffSchedule(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    pass


class OnScheduleOne(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleOne(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    class Meta(OffScheduleModelMixin.Meta):
        pass


class OnScheduleTwo(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleTwo(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    pass


class OnScheduleThree(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleThree(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    pass


class OnScheduleSix(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleSix(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    pass


class CrfOne(CrfModelMixin, BaseUuidModel):
    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    f1 = models.CharField(max_length=50, null=True, blank=True)

    f2 = models.CharField(max_length=50, null=True, blank=True)

    f3 = models.CharField(max_length=50, null=True, blank=True)

    next_appt_date = models.DateField(null=True, blank=True)

    next_visit_code = models.CharField(max_length=50, null=True, blank=True)


class CrfTwo(CrfModelMixin, BaseUuidModel):
    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    f1 = models.CharField(max_length=50, null=True, blank=True)

    f2 = models.CharField(max_length=50, null=True, blank=True)

    f3 = models.CharField(max_length=50, null=True, blank=True)

    visitschedule = models.ForeignKey(
        VisitSchedule, on_delete=PROTECT, max_length=15, null=True, blank=False
    )


class CrfThree(CrfModelMixin, BaseUuidModel):
    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    f1 = models.CharField(max_length=50, null=True, blank=True)

    f2 = models.CharField(max_length=50, null=True, blank=True)

    f3 = models.CharField(max_length=50, null=True, blank=True)

    allow_create_interim = models.BooleanField(default=False)

    appt_date = models.DateField(null=True, blank=True)


class CrfFour(CrfModelMixin, BaseUuidModel):
    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    f1 = models.CharField(max_length=50, null=True, blank=True)

    f2 = models.CharField(max_length=50, null=True, blank=True)

    f3 = models.CharField(max_length=50, null=True, blank=True)


class CrfFive(CrfModelMixin, BaseUuidModel):
    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    f1 = models.CharField(max_length=50, null=True, blank=True)

    f2 = models.CharField(max_length=50, null=True, blank=True)

    f3 = models.CharField(max_length=50, null=True, blank=True)


class NextAppointmentCrf(NextAppointmentCrfModelMixin, CrfModelMixin, BaseUuidModel):
    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)



from django.urls import include, path
from django.views.generic import RedirectView
from edc_dashboard.views import AdministrationView
from edc_subject_dashboard.views import SubjectDashboardView
from edc_utils.paths_for_urlpatterns import paths_for_urlpatterns

app_name = "edc_appointment"

urlpatterns = []

for app_name in [
    "edc_appointment",
    "edc_auth",
    "edc_device",
    "edc_protocol",
    "edc_visit_schedule",
    "edc_pharmacy",
    "edc_data_manager",
    "edc_adverse_event",
]:
    for p in paths_for_urlpatterns(app_name):
        urlpatterns.append(p)

urlpatterns.extend(
    SubjectDashboardView.urls(
        namespace=app_name,
        label="subject_dashboard",
        identifier_pattern=r"\w+",
    )
)


urlpatterns.extend(
    [
        path("administration/", AdministrationView.as_view(), name="administration_url"),
        path("i18n/", include("django.conf.urls.i18n")),
        path("", RedirectView.as_view(url="admin/"), name="home_url"),
        path("", RedirectView.as_view(url="admin/"), name="logout"),
    ]
)



from __future__ import annotations

from datetime import date, datetime
from zoneinfo import ZoneInfo

from django import forms
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from edc_metadata.utils import has_keyed_metadata
from edc_utils import convert_php_dateformat
from edc_utils.date import to_local
from edc_visit_schedule.exceptions import ScheduledVisitWindowError

from ..utils import get_appointment_by_datetime


class NextAppointmentCrfModelFormMixin:
    appt_date_fld = "appt_date"
    visit_code_fld = "visitschedule"

    def clean(self):
        cleaned_data = super().clean()
        self.validate_suggested_date_with_future_appointments()
        self.validate_suggested_visit_code()
        return cleaned_data

    @property
    def allow_create_interim(self) -> str:
        return self.cleaned_data.get("allow_create_interim", False)

    @property
    def suggested_date(self) -> date | None:
        return self.cleaned_data.get(self.appt_date_fld)

    @property
    def suggested_visit_code(self) -> str | None:
        return getattr(
            self.cleaned_data.get(self.visit_code_fld),
            "visit_code",
            self.cleaned_data.get(self.visit_code_fld),
        )

    def validate_suggested_date_with_future_appointments(self):
        if self.suggested_date and (
            self.related_visit.appointment.next.related_visit
            or has_keyed_metadata(self.related_visit.appointment.next)
        ):
            if (
                self.suggested_date
                != to_local(self.related_visit.appointment.next.appt_datetime).date()
            ):
                appointment = self.related_visit.appointment.next
                date_format = convert_php_dateformat(settings.SHORT_DATE_FORMAT)
                next_appt_date = (
                    to_local(appointment.appt_datetime).date().strftime(date_format)
                )
                raise forms.ValidationError(
                    {
                        self.appt_date_fld: _(
                            "Invalid. Next visit report already submitted. Expected "
                            "`%(dt)s`. See `%(visit_code)s`."
                        )
                        % {
                            "dt": next_appt_date,
                            "visit_code": appointment.visit_code,
                        }
                    }
                )

        if (
            self.suggested_date
            and self.suggested_date
            > to_local(self.related_visit.appointment.next.appt_datetime).date()
            and has_keyed_metadata(self.related_visit.appointment.next)
        ):
            appointment = self.related_visit.appointment.next
            date_format = convert_php_dateformat(settings.SHORT_DATE_FORMAT)
            raise forms.ValidationError(
                {
                    self.appt_date_fld: _(
                        "Invalid. Expected a date before appointment "
                        "`%(visit_code)s` on "
                        "%(dt_str)s."
                    )
                    % {
                        "visit_code": appointment.visit_code,
                        "dt_str": to_local(appointment.appt_datetime)
                        .date()
                        .strftime(date_format),
                    }
                }
            )

    def validate_suggested_visit_code(self):
        if suggested_date := self.suggested_date:
            try:
                appointment = get_appointment_by_datetime(
                    self.as_datetime(suggested_date),
                    subject_identifier=self.related_visit.subject_identifier,
                    visit_schedule_name=self.related_visit.visit_schedule.name,
                    schedule_name=self.related_visit.schedule.name,
                    raise_if_in_gap=False,
                )
            except ScheduledVisitWindowError as e:
                raise forms.ValidationError({self.appt_date_fld: str(e)})
            if not appointment:
                raise forms.ValidationError(
                    {self.appt_date_fld: _("Invalid. Must be within the followup period.")}
                )

            if (
                self.suggested_visit_code
                and self.suggested_visit_code != appointment.visit_code
            ):
                date_format = convert_php_dateformat(settings.SHORT_DATE_FORMAT)
                raise forms.ValidationError(
                    {
                        self.visit_code_fld: _(
                            "Expected %(visit_code)s using %(dt_str)s from above."
                        )
                        % {
                            "visit_code": appointment.visit_code,
                            "dt_str": suggested_date.strftime(date_format),
                        }
                    }
                )
            if appointment == self.related_visit.appointment:
                if self.allow_create_interim:
                    pass
                else:
                    raise forms.ValidationError(
                        {
                            self.appt_date_fld: (
                                _(
                                    "Invalid. Cannot be within window period "
                                    "of the current appointment."
                                )
                            )
                        }
                    )
            elif self.allow_create_interim:
                raise forms.ValidationError(
                    {
                        "allow_create_interim": _(
                            "Cannot override if date is not within window period "
                            "of the current appointment."
                        )
                    }
                )

    @staticmethod
    def as_datetime(dt: date) -> datetime:
        return datetime(dt.year, dt.month, dt.day, 23, 59, 59, tzinfo=ZoneInfo("UTC"))



from dateutil.relativedelta import relativedelta
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.visit import Visit
from edc_visit_schedule.visit_schedule import VisitSchedule

from edc_appointment_app.consents import consent_v1
from edc_appointment_app.visit_schedule.crfs import crfs, crfs_missed


def get_visit_schedule5() -> VisitSchedule:
    visit_schedule5 = VisitSchedule(
        name="visit_schedule5",
        offstudy_model="edc_appointment_app.subjectoffstudy",
        death_report_model="edc_appointment_app.deathreport",
    )

    schedule5 = Schedule(
        name="monthly_schedule",
        onschedule_model="edc_appointment_app.onschedulethree",
        offschedule_model="edc_appointment_app.offschedulethree",
        appointment_model="edc_appointment.appointment",
        consent_definitions=[consent_v1],
    )

    visits = [
        Visit(
            code="1000",
            title="Baseline",
            timepoint=0,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=0),
            add_window_gap_to_lower=True,
            requisitions=None,
            crfs=crfs,
            crfs_missed=crfs_missed,
            facility_name="7-day-clinic",
        )
    ]
    for index, visit_code in [
        (1, "1010"),
        (2, "1020"),
        (3, "1030"),
        (4, "1040"),
        (5, "1050"),
        (6, "1060"),
        (7, "1070"),
        (8, "1080"),
        (9, "1090"),
    ]:
        visits.append(
            Visit(
                code=visit_code,
                title=f"Month {index}",
                timepoint=index,
                rbase=relativedelta(months=index),
                rlower=relativedelta(days=15),
                rupper=relativedelta(days=15),
                add_window_gap_to_lower=True,
                max_window_gap_to_lower=None,
                requisitions=None,
                crfs=crfs,
                crfs_missed=crfs_missed,
                facility_name="7-day-clinic",
                allow_unscheduled=True,
            )
        )
    for visit in visits:
        schedule5.add_visit(visit)

    visit_schedule5.add_schedule(schedule5)
    return visit_schedule5



from __future__ import annotations

import datetime as dt
from zoneinfo import ZoneInfo

import time_machine
from django.db.models import ProtectedError
from django.db.models.signals import post_save
from django.test import TestCase, override_settings, tag
from edc_facility.import_holidays import import_holidays
from edc_metadata.models import CrfMetadata

from edc_appointment.constants import INCOMPLETE_APPT, NEW_APPT
from edc_appointment.managers import AppointmentDeleteError
from edc_appointment.models import Appointment
from edc_appointment.utils import delete_appointment_in_sequence, get_next_appointment
from edc_appointment_app.models import CrfOne, SubjectVisit
from edc_appointment_app.tests.appointment_app_test_case_mixin import (
    AppointmentAppTestCaseMixin,
)

from ..helper import Helper

utc_tz = ZoneInfo("UTC")


@override_settings(SITE_ID=10)
@time_machine.travel(dt.datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz))
class TestMoveAppointment(AppointmentAppTestCaseMixin, TestCase):
    helper_cls = Helper

    def setUp(self):
        super().setUp()

    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def test_appointments_resequence_after_delete(self):
        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "1000.3", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(Appointment, order_by="appt_datetime"),
        )
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=1)
        appointment.delete()
        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(Appointment, order_by="appt_datetime"),
        )

    def test_resequence_appointment_correctly2(self):
        def get_visit_codes():
            return [
                f"{o.visit_code}.{o.visit_code_sequence}"
                for o in Appointment.objects.all().order_by("appt_datetime")
            ]

        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "1000.3", "2000.0", "3000.0", "4000.0"],
            get_visit_codes(),
        )
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=2)
        appointment.delete()
        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "2000.0", "3000.0", "4000.0"],
            get_visit_codes(),
        )

    def test_resequence_appointment_correctly3(self):
        def get_visit_codes():
            return [
                f"{o.visit_code}.{o.visit_code_sequence}"
                for o in Appointment.objects.all().order_by("appt_datetime")
            ]

        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "1000.3", "2000.0", "3000.0", "4000.0"],
            get_visit_codes(),
        )
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=3)
        appointment.delete()
        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "2000.0", "3000.0", "4000.0"],
            get_visit_codes(),
        )

    def test_delete_0_appointment_in_sequence(self):
        self.assertEqual(
            [0, 1, 2, 3],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=0)
        # raises ProtectedError because Subjectvisit exists
        self.assertRaises(ProtectedError, delete_appointment_in_sequence, appointment)
        SubjectVisit.objects.get(appointment=appointment).delete()
        # raises AppointmentDeleteError (from manager) because not allowed by manager
        self.assertRaises(AppointmentDeleteError, delete_appointment_in_sequence, appointment)
        # assert nothing was done
        self.assertEqual(
            [0, 1, 2, 3],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )

    def test_delete_first_appointment_in_sequence(self):
        post_save.disconnect(dispatch_uid="appointments_on_post_delete")
        self.assertEqual(
            [0, 1, 2, 3],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=1)
        delete_appointment_in_sequence(appointment)
        self.assertEqual(
            [0, 1, 2],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )

    def test_delete_second_appointment_in_sequence(self):
        self.assertEqual(
            [0, 1, 2, 3],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=2)
        delete_appointment_in_sequence(appointment)
        self.assertEqual(
            [0, 1, 2],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )

    def test_delete_third_appointment_in_sequence(self):
        self.assertEqual(
            [0, 1, 2, 3],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=3)
        delete_appointment_in_sequence(appointment)
        self.assertEqual(
            [0, 1, 2],
            [
                o.visit_code_sequence
                for o in Appointment.objects.filter(visit_code="1000").order_by(
                    "appt_datetime"
                )
            ],
        )

    @tag("4")
    def test_appointments_and_related_visits_resequence_after_appointment_is_deleted(self):
        for appointment in Appointment.objects.all().order_by("timepoint_datetime"):
            self.create_related_visit(appointment)
        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "1000.3", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(Appointment, order_by="appt_datetime"),
        )
        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "1000.3", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(SubjectVisit, order_by="report_datetime"),
        )
        appointment_10001 = Appointment.objects.get(visit_code="1000", visit_code_sequence=1)
        appointment_10001.related_visit.delete()
        appointment_10001.delete()
        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(Appointment, order_by="appt_datetime"),
        )
        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(SubjectVisit, order_by="report_datetime"),
        )

    def test_related_visit_resequences_after_appointment_is_deleted_with_crfs_submitted(
        self,
    ):
        for appointment in Appointment.objects.all().order_by("timepoint_datetime"):
            self.create_related_visit(appointment)
        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "1000.3", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(Appointment, order_by="appt_datetime"),
        )
        # enter CRF on 1000.2
        appointment_10002 = Appointment.objects.get(visit_code="1000", visit_code_sequence=2)
        CrfOne.objects.create(subject_visit=appointment_10002.related_visit)
        CrfOne.objects.get(
            subject_visit__appointment__visit_code="1000",
            subject_visit__appointment__visit_code_sequence=2,
        )
        # delete 1000.1
        appointment_10001 = Appointment.objects.get(visit_code="1000", visit_code_sequence=1)
        appointment_10001.related_visit.delete()
        appointment_10001.delete()
        CrfOne.objects.get(
            subject_visit__appointment__visit_code="1000",
            subject_visit__appointment__visit_code_sequence=1,
        )

    def test_metadata_resequences_after_appointment_is_deleted(self):
        for appointment in Appointment.objects.all().order_by("timepoint_datetime"):
            self.create_related_visit(appointment)
        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "1000.3", "2000.0", "3000.0", "4000.0"],
            self.get_visit_codes(Appointment, order_by="appt_datetime"),
        )
        self.assertGreater(
            CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=1).count(), 0
        )
        self.assertGreater(
            CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=2).count(), 0
        )
        self.assertGreater(
            CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=3).count(), 0
        )
        # delete 1000.1
        appointment_10001 = Appointment.objects.get(visit_code="1000", visit_code_sequence=1)
        appointment_10001.related_visit.delete()
        appointment_10001.delete()

        self.assertGreater(
            CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=1).count(), 0
        )
        self.assertGreater(
            CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=2).count(), 0
        )
        self.assertEqual(
            CrfMetadata.objects.filter(visit_code="1000", visit_code_sequence=3).count(), 0
        )

    def test_appt_status_correct_after_appointment_is_deleted(self):
        for appointment in Appointment.objects.all().order_by("timepoint_datetime"):
            self.create_related_visit(appointment)
            appointment.refresh_from_db()
            self.assertEqual(appointment.appt_status, INCOMPLETE_APPT)

        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=3)
        appointment.related_visit.delete()
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=3)
        self.assertEqual(appointment.appt_status, NEW_APPT)

        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=1)
        appointment.related_visit.delete()
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=1)
        self.assertEqual(appointment.appt_status, NEW_APPT)

        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=1)
        self.assertEqual(appointment.appt_status, NEW_APPT)
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=2)
        self.assertEqual(appointment.appt_status, INCOMPLETE_APPT)
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=3)
        self.assertEqual(appointment.appt_status, NEW_APPT)

        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=1)
        appointment.delete()

        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=1)
        self.assertEqual(appointment.appt_status, INCOMPLETE_APPT)
        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=2)
        self.assertEqual(appointment.appt_status, NEW_APPT)

    def test_next_getter(self):
        def get_visit_codes():
            return [
                f"{o.visit_code}.{o.visit_code_sequence}"
                for o in Appointment.objects.all().order_by("appt_datetime")
            ]

        appointment = Appointment.objects.get(visit_code="1000", visit_code_sequence=0)

        self.assertEqual(appointment.visit_code, "1000")
        self.assertEqual(appointment.visit_code_sequence, 0)

        self.assertEqual(get_next_appointment(appointment).visit_code, "2000")
        self.assertEqual(get_next_appointment(appointment).visit_code_sequence, 0)
        self.assertEqual(
            get_next_appointment(appointment, include_interim=True).visit_code, "1000"
        )
        self.assertEqual(
            get_next_appointment(appointment, include_interim=True).visit_code_sequence, 1
        )
        self.assertEqual(
            ["1000.0", "1000.1", "1000.2", "1000.3", "2000.0", "3000.0", "4000.0"],
            get_visit_codes(),
        )



from edc_consent.consent_definition import ConsentDefinition
from edc_consent.site_consents import site_consents
from edc_constants.constants import FEMALE, MALE
from edc_protocol.research_protocol_config import ResearchProtocolConfig

consent_v1 = ConsentDefinition(
    "edc_appointment_app.subjectconsentv1",
    version="1",
    start=ResearchProtocolConfig().study_open_datetime,
    end=ResearchProtocolConfig().study_close_datetime,
    age_min=18,
    age_is_adult=18,
    age_max=64,
    gender=[MALE, FEMALE],
)

site_consents.register(consent_v1)



from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING, Any
from zoneinfo import ZoneInfo

from django.apps import apps as django_apps
from django.core.exceptions import FieldError
from django.db import IntegrityError
from edc_metadata.utils import has_keyed_metadata

from .constants import MISSED_APPT, NEW_APPT, SKIPPED_APPT
from .exceptions import AppointmentWindowError
from .models import Appointment
from .utils import (
    AppointmentAlreadyStarted,
    get_allow_skipped_appt_using,
    get_appointment_by_datetime,
    raise_on_appt_datetime_not_in_window,
    reset_appointment,
    skip_appointment,
)

if TYPE_CHECKING:
    from django.db.models import QuerySet


class AnyCRF(Any):
    pass


class SkipAppointmentsError(Exception):
    pass


class SkipAppointmentsValueError(Exception):
    pass


class SkipAppointmentsFieldError(Exception):
    pass


class SkipAppointments:
    """Using a future date from a CRF, update the `appt_datetime` of
    the appointment that falls within the window period of the date
    AND set the `app_status` of interim appointments to `SKIPPED_APPT`.

    * CRF has a datefield which captures the date when the patient
      is next expected;
    * CRF has a charfield which captures the next visit code that is
      within the window period of the date.
    * You should validate the next visit code and the date before
      calling (e.g. on the form).
    """

    def __init__(self, crf_obj: AnyCRF):
        self._last_crf_obj = None
        self._next_appt_date: date | None = None
        self._next_appt_datetime: datetime | None = None
        self._next_visit_code: str | None = None
        self._scheduled_appointments: QuerySet[Appointment] | None = None
        if not get_allow_skipped_appt_using().get(crf_obj._meta.label_lower):
            raise SkipAppointmentsError(
                "Appointments may not be skipped. "
                "settings.EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING="
                f"`{get_allow_skipped_appt_using()}`"
                f"Got model `{crf_obj._meta.label_lower}`."
            )
        self.crf_obj = crf_obj
        self.dt_fld, self.visit_code_fld = get_allow_skipped_appt_using().get(
            self.crf_obj._meta.label_lower
        )
        self.crf_model_cls = django_apps.get_model(self.crf_obj._meta.label_lower)
        self.related_visit_model_attr: str = self.crf_obj.related_visit_model_attr()
        self.appointment: Appointment = getattr(
            self.crf_obj, self.related_visit_model_attr
        ).appointment
        self.subject_identifier: str = self.appointment.subject_identifier
        self.visit_schedule_name: str = self.appointment.visit_schedule_name
        self.schedule_name: str = self.appointment.schedule_name

    def update(self) -> bool:
        """Reset appointments and set any as skipped up to the
        date provided from the CRF.

        Return True if next scheduled appointment is updated.
        """
        self.reset_appointments()
        return self.update_appointments()

    def reset_appointments(self):
        """Reset any Appointments previously where `appt_status`
        is SKIPPED_APPT.

        Also reset `appt_datetime` on any new appts (NEW_APPT).
        """
        # reset auto-created MISSED_APPT (appt_type__isnull=True)
        # TODO: this for-loop block may be removed once all auto-created
        #    missed appts are removed.
        for appointment in self.scheduled_appointments.filter(
            appt_type__isnull=True, appt_timing=MISSED_APPT, visit_code_sequence=0
        ).exclude(
            appt_status__in=[SKIPPED_APPT, NEW_APPT],
        ):
            try:
                reset_appointment(appointment)
            except AppointmentAlreadyStarted:
                pass

        for appointment in self.scheduled_appointments.filter(
            appt_datetime__gt=self.appointment.appt_datetime,
        ):
            try:
                reset_appointment(appointment)
            except IntegrityError as e:
                print(e)
            except AppointmentAlreadyStarted:
                pass

    def update_appointments(self) -> bool:
        """Return True if next scheduled appointment is updated.

        Update Appointments up the next apointment date.

        Set `appt_status` = SKIPPED_APPT up the appointment BEFORE
        the date from the CRF.

        Only set the `appt_datetime` for the appointment where the  date
        from the CRF lands in the window period. Leave status as NEW_APPT.

        Stop if any appointment has keyed data.
        """
        next_scheduled_appointment_updated = False
        skip_comment = (
            f"based on date reported at {self.last_crf_obj.related_visit.visit_code}"
        )

        cancelled_appointments = []
        appointment = self.appointment
        while appointment:
            if appointment.visit_code_sequence > 0:
                if appointment.appt_status == NEW_APPT:
                    cancelled_appointments.append(appointment)
            elif self.is_next_scheduled(appointment):
                try:
                    self.update_appointment_as_next_scheduled(appointment)
                except AppointmentAlreadyStarted:
                    pass
                else:
                    next_scheduled_appointment_updated = True
                break
            else:
                try:
                    skip_appointment(appointment, comment=skip_comment)
                except AppointmentAlreadyStarted:
                    pass
            appointment = appointment.relative_next
        for appointment in cancelled_appointments:
            appointment.delete()
        return next_scheduled_appointment_updated

    def is_next_scheduled(self, appointment):
        return (
            appointment.visit_code == self.next_visit_code
            and appointment.visit_code_sequence == 0
        )

    def update_appointment_as_next_scheduled(self, appointment: Appointment) -> None:
        """Return True if this is the "next" appointment (the last
        appointment in the sequence).

        If "next", try to update if CRfs not yet submitted/KEYED.
        """
        if has_keyed_metadata(appointment) or appointment.related_visit:
            raise AppointmentAlreadyStarted(
                f"Unable update as next. Appointment already started. Got {appointment}."
            )
        else:
            appointment.appt_status = NEW_APPT
            appointment.appt_datetime = self.next_appt_datetime
            appointment.comment = ""
            self.validate_appointment_as_next(appointment)
            appointment.save(update_fields=["appt_status", "appt_datetime", "comment"])

    @property
    def last_crf_obj(self):
        """Return the CRF instance for the last timepoint /
        report_datetime.
        """
        if not self._last_crf_obj:
            try:
                self._last_crf_obj = (
                    self.crf_model_cls.objects.filter(**self.query_opts)
                    .order_by(f"{self.related_visit_model_attr}__report_datetime")
                    .last()
                )
            except FieldError as e:
                raise SkipAppointmentsFieldError(
                    f"{e}. See {self.crf_model_cls._meta.label_lower}."
                )
        return self._last_crf_obj

    @property
    def query_opts(self) -> dict[str, str]:
        return {
            f"{self.related_visit_model_attr}__subject_identifier": self.subject_identifier,
            f"{self.related_visit_model_attr}__visit_schedule_name": self.visit_schedule_name,
            f"{self.related_visit_model_attr}__schedule_name": self.schedule_name,
        }

    @property
    def next_appt_date(self) -> date | None:
        """Return the date from the CRF."""
        if not self._next_appt_date:
            try:
                self._next_appt_date = getattr(self.last_crf_obj, self.dt_fld)
            except AttributeError:
                raise SkipAppointmentsFieldError(
                    f"Unknown field name for next scheduled appointment date. See "
                    f"{self.last_crf_obj._meta.label_lower}. Got `{self.dt_fld}`."
                )
        return self._next_appt_date

    @property
    def next_appt_datetime(self) -> datetime:
        """Return a datetime representation of next_appt_date."""
        if not self._next_appt_datetime:
            self._next_appt_datetime = datetime(
                year=self.next_appt_date.year,
                month=self.next_appt_date.month,
                day=self.next_appt_date.day,
                hour=6,
                minute=0,
                tzinfo=ZoneInfo("UTC"),
            )
        return self._next_appt_datetime

    @property
    def scheduled_appointments(self) -> QuerySet[Appointment]:
        """Return a queryset of scheduled appointments for this
        subject's schedule (visit_code_sequence=0).
        """
        if not self._scheduled_appointments:
            self._scheduled_appointments = Appointment.objects.filter(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
                visit_code_sequence=0,
            ).order_by("timepoint_datetime")
        return self._scheduled_appointments

    @property
    def visit_codes(self) -> list[str]:
        """Return a list of scheduled visit codes for this subject's
        schedule.
        """
        return [obj.visit_code for obj in self.scheduled_appointments]

    @property
    def next_visit_code(self) -> str:
        """Return the suggested visit_code entered on the last
        CRF instance ir raises.
        """
        if not self._next_visit_code:
            try:
                self._next_visit_code = getattr(self.last_crf_obj, self.visit_code_fld)
            except AttributeError:
                raise SkipAppointmentsFieldError(
                    "Unknown field name for visit code. See "
                    f"{self.last_crf_obj._meta.label_lower}. Got `{self.visit_code_fld}`."
                )
            self._next_visit_code = getattr(
                self._next_visit_code, "visit_code", self._next_visit_code
            )
            if self._next_visit_code not in self.visit_codes:
                raise SkipAppointmentsValueError(
                    "Invalid value for visit code. Expected one of "
                    f"{','.join(self.visit_codes)}. See {self.last_crf_obj._meta.label_lower}."
                    f"{self.visit_code_fld}`. Got `{self._next_visit_code}`"
                )
        return self._next_visit_code

    def validate_appointment_as_next(self, appointment):
        try:
            raise_on_appt_datetime_not_in_window(appointment)
        except AppointmentWindowError as e:
            raise SkipAppointmentsValueError(e)

        next_appt = get_appointment_by_datetime(
            self.next_appt_datetime,
            appointment.subject_identifier,
            appointment.visit_schedule_name,
            appointment.schedule_name,
            raise_if_in_gap=False,
        )
        if next_appt.visit_code != self.next_visit_code:
            raise SkipAppointmentsValueError(
                "Visit code is invalid for appointment datetime. "
                f"You suggested {self.next_visit_code}. "
                f"The appointment datetime matches with {next_appt.visit_code}."
            )

        next_appointment = get_appointment_by_datetime(
            self.next_appt_datetime,
            appointment.subject_identifier,
            appointment.visit_schedule_name,
            appointment.schedule_name,
            raise_if_in_gap=False,
        )
        return self.appointment == next_appointment

        # and getattr(
        #     self.crf_obj, "allow_create_interim", None
        # ):
        #         subject_identifier=self.appointment.subject_identifier,
        #         visit_schedule_name=self.appointment.visit_schedule_name,
        #         schedule_name=self.appointment.schedule_name,
        #         timepoint=self.appointment.timepoint,
        #         visit_code=self.appointment.visit_code,
        #         visit_code_sequence=self.appointment.visit_code_sequence + 1,
        #         facility=self.appointment.facility,
        #     )
